#!/usr/bin/env python
# ===================================================================
# ScriptName:   gui_map
# Purpose:      User interface utility to handle maps and segmentations
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/09
# Revision:     v1.2
# Last Change:  2024/11/19: added loadTif
#               2024/08/09: separated from gui.py 
# ===================================================================

import os
import time
import shutil
from pathlib import Path
import numpy as np
from PIL import Image
Image.MAX_IMAGE_PIXELS = None
import mrcfile
import tifffile
import subprocess
from skimage import exposure, transform
import dearpygui.dearpygui as dpg

from SPACEtomo.modules.gui.gui import showInfoBox
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log

# Check if IMOD is available (needed for mrc file import)
if "IMOD" in os.environ["PATH"] or shutil.which("newstack") is not None:
    IMOD = True
else:
    IMOD = False

class MMap:
    def __init__(self, file_path, pix_size=1, stitched=True, tile_size=(1024, 1024), quantile=0.01, status=None) -> None:
        self.file = Path(file_path)
        self.pix_size = pix_size
        self.stitched = stitched
        self.tile_size = np.array(tile_size)
        self.status = status # for GUI status line update
        self.img = np.zeros((1, 1))
        self.img_bin = self.img
        self.img_mod = None # Option to hold modified version of image
        self.binning = 1

        if self.file.suffix.lower() in [".mrc", ".map"]:
            if self.loadMrc(quantile=quantile):
                return

        if self.file.suffix.lower() in [".png", ".tif"]:
            start = time.time()
            if self.file.suffix.lower() == ".tif":
                if not self.loadTif():
                    return
            else:
                # Catch truncated files
                try:
                    #self.img = np.array(Image.open(self.file)).astype(float) / 255
                    self.img = utils.toNumpy(Image.open(self.file))
                except OSError:
                    log(f"ERROR: Image file [{self.file}] did not finish saving. Please try again later or make sure the file was not corrupted!")
                    return
            log(f"DEBUG: Loaded map in {time.time() - start} s. [{self.img.dtype}]")

        # Fill also when loadMrc fails
        self.pix_size = pix_size        # Angstrom / px
        self.tile_num = (self.img.shape[:2] / self.tile_size).astype(int)
        self.calculateTileBounds()

        # Dims in microns
        self.dims_microns = np.flip(self.img.shape[:2]) * self.pix_size / 10000

    def loadMrc(self, quantile):
        """Loads mrc file and assembles montage. (TODO: outsource to Mrc class)"""

        arg_pix_size = self.pix_size
        arg_tile_size = self.tile_size

        # If IMOD is available: extract tile coordinates from header
        if IMOD:
            pieces_file = self.file.parent / (self.file.stem + "_pieces.txt")
            subprocess.run(["extractpieces", "-input", self.file, "-output", pieces_file], stdout=subprocess.DEVNULL)
            piece_data = np.array(utils.loadColFile(pieces_file, int))
            if len(piece_data) > 0:
                # Delete file
                pieces_file.unlink()

                # Check if file contains more than 1 montage
                if piece_data.shape[1] > 2:
                    unique = np.unique(piece_data[:, 2])
                    if len(unique) > 1:
                        log(f"WARNING: File contains {len(unique)} montage maps. Use IMOD's edmont command to split the file!")
                        showInfoBox("WARNING", f"File contains {len(unique)} montage maps.\nDo you want to run IMOD's edmont to split the file by montage?", splitMontageFile, ["Yes", "No"], [self.file, ""])
                        self.img = np.zeros((1, 1))
                        return
        else:
            piece_data = []

        # Load mrc file
        with mrcfile.open(self.file, permissive=True) as mrc:
            # Get pixel data
            data = mrc.data
            self.pix_size = float(mrc.voxel_size.x) #[A]
        
            # Add third dimension in case of single image
            if data.ndim < 3:
                data = np.expand_dims(data, 0)

            # Get size of single tile
            self.tile_size = np.array(data.shape[1:3], dtype=int)

            # Pre-bin mrc data x2 in case pixel size is <4x model size and large montages do speed up scaling
            # Thresholds of 1/4 model pixel size and 10 montage pieces were chosen arbitrarily
            prebinned = False
            if self.pix_size * 3 < arg_pix_size and data.shape[0] > 10:
                data = data[:, ::2, ::2]
                self.pix_size = self.pix_size * 2
                self.tile_size = self.tile_size // 2
                prebinned = True
                log("WARNING: Map was prebinned by 2 to speed up processing.")

            # Figure out scaling based on quantiles
            vals = np.round(np.array((np.min(data), np.max(data), np.mean(data)), dtype=np.float32), 2)
            if self.status is not None:
                self.status.update("Scaling mrc file...", box=True)
            log("Loading mrc file...")
            log("# Map statistics:")
            log(f"# Min, max, mean: {vals}")
            log(f"# Quantile: {round(quantile, 5)}")
            log(f"# Cutoffs: {np.quantile(data, quantile)}, {np.round(np.quantile(data, 1 - quantile), 2)}")
            data = exposure.rescale_intensity(data, in_range=(np.quantile(data, quantile), np.quantile(data, 1 - quantile)), out_range=(0, 1))

            if self.status is not None:
                self.status.update("Assembling tiles...", box=True)
            log("Assembling tiles...")
            montage_nums = []

            # If IMOD is available: use tile coordinates
            if len(piece_data) > 0:
                log("Using extracted piece coordinates...")
                tile_coords = []
                self.tile_bounds = [[], []]
                for piece in piece_data:
                    tile_coords.append([piece[1], piece[0]])
                    # Add coords also to tile bounds
                    #if piece[0] not in self.tile_bounds[0]:
                    #    self.tile_bounds[0].append(piece[0])
                    #if piece[1] not in self.tile_bounds[1]:
                    #    self.tile_bounds[1].append(piece[1])    

                tile_coords = np.array(tile_coords)

                if prebinned:
                    tile_coords = tile_coords // 2

                # Figure out montage dimensions from coordinates
                unique_y, counts = np.unique(tile_coords[:, 0], return_counts=True)
                self.tile_num = np.array([counts[-1], data.shape[0] / counts[-1]], dtype=int)

                # Get tile_bounds by extracting unique coords
                unique_x = np.unique(tile_coords[:, 1])
                self.tile_bounds = [unique_x, unique_y]

                # Assemble image
                self.img = np.zeros((np.max(tile_coords[:, 0]) + self.tile_size[0], np.max(tile_coords[:, 1]) + self.tile_size[1])) #self.tile_num * self.tile_size)
                for t, tile in enumerate(data):
                    # Stop when end of tile_coords is reached (in case there are more montages in file)
                    if t >= len(tile_coords): 
                        break
                    self.img[tile_coords[t][0]: tile_coords[t][0] + self.tile_size[0], tile_coords[t][1]: tile_coords[t][1] + self.tile_size[1]] = tile

                #else:
                #    print("WARNING: Loaded file is no montage!")
                #    self.img = data[0]
                #    self.tile_num = np.array([1, 1], dtype=int)
                #    self.calculateTileBounds()

                self.stitched = True
                #self.tile_size = arg_tile_size
                #self.calculateTileBounds()

            # If IMOD is not available: guess montage dimensions
            else:
                self.tile_num = utils.guessMontageDims(data.shape[0])
                if self.tile_size[0] < self.tile_size[1]:
                    self.tile_num = np.flip(self.tile_num)

                if not IMOD:
                    if self.status is not None:
                        self.status.update()
                    log(f"WARNING: No IMOD installation found! Cannot access montage tile coordinates. Best guess for montage dimensions is: {self.tile_num[0]}x{self.tile_num[1]}")
                    showInfoBox("WARNING", f"No IMOD installation found! Cannot access montage tile coordinates.\nBest guess for montage dimensions is: {self.tile_num[0]}x{self.tile_num[1]}\nPlease make sure IMOD is on the PATH to get better results!")
                else:
                    log(f"WARNING: No piece coordinates found. Best guess for montage dimensions is: {self.tile_num[0]}x{self.tile_num[1]}")

                # Assemble image
                self.img = np.zeros(self.tile_num * self.tile_size)
                for t, tile in enumerate(data):
                    x = t % self.tile_num[0] * self.tile_size[0]
                    y = t // self.tile_num[0] * self.tile_size[1]

                    # Account for serpentine acquisition (at least for WG maps)
                    if (t // self.tile_num[0]) % 2 == 1:
                        x = self.img.shape[0] - x - self.tile_size[0]

                    self.img[x: x + self.tile_size[0], y: y + self.tile_size[1]] = tile

                if self.stitched:
                    log("WARNING: Map was not stitched. Using tile images as tiles!")
                self.stitched = False

                self.calculateTileBounds()
     
            # Flip y
            self.img = np.flip(self.img, axis=0)

            #self.pix_size = float(mrc.voxel_size.x) #[A]

            return True

    def loadTif(self):
        """Loads tif file. TODO: outsource to allow for multiple MMaps from single file"""

        img = tifffile.imread(self.file)

        # Check channels
        if img.ndim > 3:                # ZCXY format
            log(f"ERROR: Loading of multi-channel stacks is currently not supported. Please provide a single image per channel!")
            return False
        elif img.ndim > 2:              # CXY or XYC format
            if img.shape[0] <= 10:      # CXY
                log(f"DEBUG: TIF: CXY with {img.shape[0]} channels")
                log(f"WARNING: TIF file contains multiple channels. Channels have been split into separate files and only first channel was loaded. Please load the other channels manually!")
                for c, channel in enumerate(img):
                    channel_rescaled = exposure.rescale_intensity(channel, out_range=(0, 255))
                    if c == 0:
                        self.img = channel_rescaled
                    Image.fromarray(channel_rescaled.astype(np.uint8)).save(self.file.parent / (self.file.stem + f"_ch{c}.tif"))

            elif img.shape[2] <= 10:    # XYC
                log(f"DEBUG: TIF: XYC with {img.shape[2]} channels")
                self.img = img
        else:                           # XY format
            log(f"DEBUG: TIF: Single channel tif")
            self.img = exposure.rescale_intensity(img, out_range=(0, 255))
        return True

    def checkBinning(self, default_binning=1):
        """Checks if map has to be binned."""

        start = time.time()
        self.binning = default_binning
        dims = np.flip(np.array(self.img.shape[:2])) // self.binning
        while np.max(dims) > 16384:  # hard limit for texture sizes on apple GPU
            self.binning += 1
            dims = np.flip(np.array(self.img.shape[:2])) // self.binning

        if self.binning > 1:
            log("WARNING: Map is too large and will be binned by " + str(self.binning) + " (for display only)! Export will be unbinned.")
            self.img_bin = self.img[::self.binning, ::self.binning]
            #dims = np.flip(np.array(self.img_bin.shape))
        else:
            self.img_bin = self.img
        log(f"Map binning took {time.time() - start} s.")

    def padDims(self, dims, transparency=True):
        """Adds channel dimensions to img."""

        if self.img_bin.ndim < 3:
            self.img_bin = np.dstack([self.img_bin] * dims)
        elif self.img_bin.shape[2] < 3:
            self.img_bin = np.dstack([self.img_bin[:, :, 0]] * dims)
        elif self.img_bin.shape[2] < 4 or transparency:
            self.img_bin = np.dstack([self.img_bin[:, :, 0], self.img_bin[:, :, 1], self.img_bin[:, :, 2], np.average(self.img_bin[:, :, :3], axis=2)])

        if not transparency:
            self.img_bin[:, :, 3] = np.full(self.img_bin.shape[:2], 255 if np.max(self.img_bin) > 1 else 1)

    def getTexture(self, mod=False):
        """Creates texture of map."""

        if not mod:
            img = self.img_bin
        elif self.img_mod is not None:
            img = self.img_mod
        else:
            log(f"ERROR: Modified image not available!")
            return np.zeros(1)

        start = time.time()
        # Convert to float
        if img.dtype == np.uint8 or np.max(img) > 1:
            img = img.astype(np.float32) / 255
        
        # Create RGBA img and flatten
        if np.ndim(img) == 3 and img.shape[2] > 3:
            # Image with transparency: just flatten
            texture_img = np.ravel(img)
        elif np.ndim(img) == 3 and img.shape[2] == 3:
            # Image with RGB: keep channels and add ones for alpha
            texture_img = np.ravel(np.dstack([img[:, :, :3], np.ones(img.shape[:2])]))
        elif np.ndim(img) == 3 and img.shape[2] < 3:
            # Image with less than 3 channels but 3 dims
            texture_img = np.ravel(np.dstack([img[:, :, :1]] * 3 + [np.ones(img.shape[:2])]))
        else:
            # Image as 2D array without 3rd dim
            texture_img = np.ravel(np.dstack([img] * 3 + [np.ones(img.shape[:2])]))
        log(f"Map texture was created in {time.time() - start} s.")

        # Get dims for texture
        dims = np.flip(np.array(img.shape[:2])).tolist()

        start = time.time()
        # Load texture
        with dpg.texture_registry():
            texture = dpg.add_static_texture(width=dims[0], height=dims[1], default_value=texture_img)
        log(f"DEBUG: Map texture was loaded in {time.time() - start} s.")

        return texture
    
    def getTileTextures(self, split=(4, 4)):
        """Creates tiled textures of map instead of single texture. (Speeds up loading, but slows unloading in case of too many tiles, could also be used for plotting unbinned map.)"""

        start = time.time()
        # Convert to float
        if self.img_bin.dtype == np.uint8:
            self.img_bin = self.img_bin.astype(np.float32) / 255
        
        # Get tile size and crop img to be divisible by tile size
        tile_size = np.array(self.img_bin.shape) // np.array(split)
        self.img_bin = self.img_bin[0: tile_size[0] * split[0], 0: tile_size[1] * split[1]]

        # Get dims for texture
        dims = np.flip(np.array(tile_size)).tolist()

        # Loop over all tiles
        textures = []
        bounds = []
        for x, v in enumerate(np.vsplit(self.img_bin, split[0])):
            for y, h in enumerate(np.hsplit(v, split[1])):
                # Create RGBA image and flatten
                texture_img = np.ravel(np.dstack([h] * 3 + [np.ones(h.shape)]))

                # Get tile bounds
                bound_min = np.array([self.img_bin.shape[0] - (x + 1) * tile_size[0], y * tile_size[1]])
                bound_max = bound_min + tile_size
                bounds.append([[bound_min[1], bound_max[1]], [bound_min[0], bound_max[0]]])

                # Load texture
                with dpg.texture_registry():
                    textures.append(dpg.add_static_texture(width=dims[0], height=dims[1], default_value=texture_img))

        log(f"Map texture was created in {time.time() - start} s.")
        
        bounds = np.array(bounds) * self.pix_size / 10000 * self.binning
        return textures, bounds

    def rescale(self, rescale_pix_size, padding=(0, 0)):
        """Rescales map to specific pixel size."""

        rescale_factor = self.pix_size / rescale_pix_size
        # Don't rescale if factor is less than 1%
        if abs(rescale_factor - 1) <= 0.01 and self.img.shape[0] > padding[0] and self.img.shape[1] > padding[1]:
            log("ERROR: New pixel size is unchanged. No rescaling was performed.")
            showInfoBox("ERROR", "New pixel size is unchanged. No rescaling was performed.")
            return
        
        if self.status is not None:
            self.status.update("Rescaling map...", box=True)
        
        self.img = transform.rescale(self.img, rescale_factor)

        self.pix_size = rescale_pix_size

        # Padding (optional)
        if self.img.shape[0] < padding[0] or self.img.shape[1] < padding[1]:
            img_pad = np.zeros([max(self.img.shape[0], padding[0]), max(self.img.shape[1], padding[1])])
            img_pad[(img_pad.shape[0] - self.img.shape[0]) // 2: (img_pad.shape[0] + self.img.shape[0]) // 2, (img_pad.shape[1] - self.img.shape[1]) // 2: (img_pad.shape[1] + self.img.shape[1]) // 2] = self.img
            self.img = img_pad
            self.tile_size = padding
            self.tile_num = np.array([1, 1], dtype=int)
        else:
            if not self.stitched:
                # Crop image for even tiles
                self.img = self.img[0:self.img.shape[0] - self.img.shape[0] % self.tile_num[0], 0:self.img.shape[1] - self.img.shape[1] % self.tile_num[1]]
                self.tile_size = (np.array(self.img.shape) / self.tile_num).astype(int)
            else:
                self.tile_num = (np.array(self.img.shape) / self.tile_size).astype(int)

        self.calculateTileBounds()

    def stitch(self):
        if self.status is not None:
            self.status.update("Stitching map...", box=True)
        # Stitch map based on cross correlation of overlaps
        # TODO
        pass

    def calculateTileBounds(self):
        self.tile_bounds = [[], []]
        # x
        for i in range(self.tile_num[1] + 1):
            self.tile_bounds[0].append(i * self.tile_size[1])
        # y
        for i in range(self.tile_num[0] + 1):
            self.tile_bounds[1].append(i * self.tile_size[0])

    def changeTiling(self, tile_size):
        self.tile_size = tile_size
        self.tile_num = (self.img.shape / self.tile_size).astype(int)
        self.calculateTileBounds()

    def returnTile(self, i, j, padding=True, padding_threshold=0.8):
        bounds = [[i * self.tile_size[0], (i + 1) * self.tile_size[0]], [j * self.tile_size[1], (j + 1) * self.tile_size[1]]]

        #print(i, j, ":", (bounds[0][1] - self.img.shape[0]) / self.tile_size[0], (bounds[1][1] - self.img.shape[1]) / self.tile_size[1])
        # If lower bounds are out of image: return False
        if bounds[0][0] < 0 or bounds[0][0] > self.img.shape[0] or bounds[1][0] < 0 or bounds[1][0] > self.img.shape[1]:
            return False, np.zeros([2, 2])
        # If larger bounds are out of image: return padded tile (if tile itself is larger than padding_threshold of normal tile)
        elif (bounds[0][1] > self.img.shape[0] or bounds[1][1] > self.img.shape[1]) and padding:
            if padding_threshold > 0 and (bounds[0][1] - self.img.shape[0] > padding_threshold * self.tile_size[0] or bounds[1][1] - self.img.shape[1] > padding_threshold * self.tile_size[1]):
                return False, np.zeros([2, 2])
            tile = np.zeros(self.tile_size)
            tile[0: min([self.tile_size[0], self.tile_size[0] - bounds[0][1] + self.img.shape[0]]), 0: min([self.tile_size[1], self.tile_size[1] - bounds[1][1] + self.img.shape[1]])] = self.img[bounds[0][0]: bounds[0][1], bounds[1][0]: bounds[1][1]]
            return tile, bounds
        # If all bounds within image: return tile
        else:
            return self.img[bounds[0][0]: bounds[0][1], bounds[1][0]: bounds[1][1]], bounds
        
    def px2microns(self, coords_px):
        """Converts img coordinates to plot coordinates."""

        return np.array([coords_px[1] * self.pix_size / 10000, self.dims_microns[1] - coords_px[0] * self.pix_size / 10000])

    def microns2px(self, coords_micron):
        """Converts plot coordinates to image coordinates."""

        return np.array([self.img.shape[0] - coords_micron[1] / self.pix_size * 10000, coords_micron[0] / self.pix_size * 10000])
    

class Segmentation:
    def __init__(self, file_path, model, pix_size=1) -> None:
        self.file = Path(file_path)
        # Check if file is .png file
        if file_path.suffix.lower() in [".png"]:
            start = time.time()
            self.img = utils.toNumpy(Image.open(self.file))
            log(f"DEBUG: Segmentation was loaded in {time.time() - start} s. [{self.img.dtype}]")
            self.valid = True
        else:
            showInfoBox("ERROR", "Can only read segmentations from .png files!")
            self.valid = False
            return
        # Check if segmentation contains data (manual_selection creates empty segmentation)
        if not np.any(self.img):
            self.valid = False
            return
        # Check if image has to be binned
        self.pix_size = pix_size
        self.checkBinning()
        # Get model for categories
        self.model = model

    # Bin map if hard threshold is surpassed
    def checkBinning(self, default_binning=1):
        self.binning = default_binning
        dims = np.flip(np.array(self.img.shape)) // self.binning
        while np.max(dims) > 16384:  # hard limit for texture sizes on apple GPU
            self.binning += 1
            dims = np.flip(np.array(self.img.shape)) // self.binning

        if self.binning > 1:
            self.img_bin = self.img[::self.binning, ::self.binning]
        else:
            self.img_bin = self.img

    # Create binary mask from selected classes
    def getMask(self, cats, unbinned=False):
        if unbinned:
            img = self.img
        else:
            img = self.img_bin

        mask = np.zeros(img.shape)
        for name in cats:
            mask[img == self.model.categories[name]] = 1

        return mask
    
    def getMaskTexture(self, cats, color=(255, 0, 0, 64)):

        # Get binary mask from classes
        start = time.time()
        mask = self.getMask(cats)
        log(f"Created mask in {time.time() - start} s.")

        # Get color image for texture
        start = time.time()
        texture_img = np.ravel(np.dstack([mask] * 3 + [np.ones(mask.shape)]) * (np.array(color) / 255))
        log(f"Created texture in {time.time() - start} s.")

        # Get dims
        dims = np.flip(np.array(mask.shape)).tolist()   # texture requires list, not numpy

        # Make texture
        start = time.time()
        with dpg.texture_registry():
            texture = dpg.add_static_texture(width=dims[0], height=dims[1], default_value=texture_img)
        log(f"DEBUG: Loaded texture in {time.time() - start} s.")
        return texture

    def getMaskTileTextures(self, cats, color=(255, 0, 0, 64), split=(4, 4)):
        """Creates tiled textures of map instead of single texture. (Speeds up loading, but slows unloading in case of too many tiles, could also be used for plotting unbinned map.)"""

        # Get binary mask from classes
        start = time.time()
        mask = self.getMask(cats)
        log(f"Created mask in {time.time() - start} s.")

        start = time.time()
        # Get tile size and crop img to be divisible by tile size
        tile_size = np.array(mask.shape) // np.array(split)
        mask = mask[0: tile_size[0] * split[0], 0: tile_size[1] * split[1]]

        # Get dims and color for texture
        dims = np.flip(np.array(tile_size)).tolist()
        color = np.array(color) / 255

        # Loop over all tiles
        textures = []
        bounds = []
        for x, v in enumerate(np.vsplit(mask, split[0])):
            for y, h in enumerate(np.hsplit(v, split[1])):
                # Create RGBA image and flatten
                texture_img = np.ravel(np.dstack([h] * 3 + [np.ones(h.shape)]) * color)

                # Get tile bounds
                bound_min = np.array([mask.shape[0] - (x + 1) * tile_size[0], y * tile_size[1]])
                bound_max = bound_min + tile_size
                bounds.append([[bound_min[1], bound_max[1]], [bound_min[0], bound_max[0]]])

                # Load texture
                with dpg.texture_registry():
                    textures.append(dpg.add_static_texture(width=dims[0], height=dims[1], default_value=texture_img))

        log(f"DEBUG: Loaded texture in {time.time() - start} s.")
        
        bounds = np.array(bounds) * self.pix_size / 10000 * self.binning
        return textures, bounds



# TODO: make file for mrc processing functions?
def splitMontageFile(sender, app_data, user_data):
    if user_data is not None and dpg.does_item_exist(user_data[0]):
        dpg.delete_item(user_data[0])
        dpg.split_frame()

    file = Path(user_data[1])    
    if len(user_data) > 1 and file.exists() and IMOD:
        processing_box = showInfoBox("PROCESSING", "Splitting mrc file into montages...", options=None)

        # Extract pieces
        pieces_file = file.parent / (file.stem + "_pieces.txt")
        subprocess.run(["extractpieces", "-input", file, "-output", pieces_file], stdout=subprocess.DEVNULL)
        if pieces_file.exists():
            with open(pieces_file, "r") as f:
                pieces = f.readlines()
            tile_nums = []
            tile_coords = []
            for piece in pieces:
                col = piece.strip().split()
                # If end of first montage is reached, reset tile_coords
                if len(col) > 2:
                    if int(col[2]) > len(tile_nums):
                        tile_nums.append(len(tile_coords))
                        tile_coords = []
                tile_coords.append([int(col[1]), int(col[0])])
            # Add final montage
            tile_nums.append(len(tile_coords))

            # If more than 1 montage, run edmont command
            if len(tile_nums) > 1:
                cmd = ["edmont", "-imin", file]
                for i in range(len(tile_nums)):
                    cmd = cmd + ["-imout", file.parent / (file.stem + "_" + str(i + 1) + file.suffix)]
                subprocess.run(cmd, stdout=subprocess.DEVNULL)

                # Clean up
                #if not os.path.exists(os.path.join(os.path.dirname(file), "multi_montage_files")): 
                if not (file.parent / "multi_montage_files").exists():
                    (file.parent / "multi_montage_files").mkdir()
                    #os.makedirs(os.path.join(os.path.dirname(file), "multi_montage_files"))
                #shutil.move(file, os.path.join(os.path.dirname(file), "multi_montage_files", os.path.basename(file)))
                shutil.move(file, file.parent / "multi_montage_files" / file.name)

        dpg.delete_item(processing_box)
        dpg.split_frame()

        log(f"{file.name} was split into {len(tile_nums)} files!")
        showInfoBox("INFO", f"{file.name} was split into {len(tile_nums)} montage files.\nPlease try loading one of them!")


#!/usr/bin/env python
# ===================================================================
# ScriptName:   thmb
# Purpose:      User interface utility to generate window with thumbnails for map selection
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/07
# Revision:     v1.2
# Last Change:  2024/08/07: separated from old SPACEtomo_tgt.py 
# ===================================================================

import os
os.environ["__GLVND_DISALLOW_PATCHING"] = "1"           # helps to minimize Segmentation fault crashes on Linux when deleting textures
import sys
import time
from pathlib import Path
try:
    import dearpygui.dearpygui as dpg
except:
    print("ERROR: DearPyGUI module not installed! If you cannot install it, please run the GUI from an external machine.")
    sys.exit()
from PIL import Image
Image.MAX_IMAGE_PIXELS = None
import numpy as np

from SPACEtomo.modules.gui import gui
from SPACEtomo.modules.utils import log

class MapWindow:
    """Windows showing thumbnails of all lamella maps in map list."""

    def __init__(self, cur_dir, map_name, map_list, map_list_tgtnum, callback_selectMap, use_thumbnails=False, thumbnail_size=(100, 100)) -> None:
        self.cur_dir = Path(cur_dir)
        self.map_name = map_name
        self.map_list = map_list
        self.map_list_tgtnum = map_list_tgtnum
        self.callback_selectMap = callback_selectMap
        self.use_thumbnails = use_thumbnails
        self.thumbnail_size = thumbnail_size

        self.map_window = dpg.add_window(label="Lamella maps", tag="map_window", no_collapse=True, popup=True, show=False)
        self.map_table = None
        self.map_labels = {}
        self.map_buttons = {}
        self.map_textures = {}
        self.generic_texture = None
        self.texture_registry = dpg.add_texture_registry()

        self.loadThumbnails(first=True)

    def makeMapTable(self, grid_cols=5):
        """Generates table with thumbnails."""

        if self.map_table is None:
            with dpg.group(tag="map_stats", parent=self.map_window):
                dpg.add_text(f"Choose a map to load!", color=gui.COLORS["heading"])
                dpg.add_text(f"Number of maps:    {len(self.map_list)}\nNumber of targets: {np.sum(self.map_list_tgtnum)}")

            # Add button to generate thumbnails
            if not self.use_thumbnails:
                dpg.add_button(label="Generate thumbnails", tag="map_butthumb", callback=lambda: self.loadThumbnails(self.thumbnail_size, override=True), parent=self.map_window)

            with dpg.table(header_row=False, parent=self.map_window) as self.map_table:
                # Add table columns
                for j in range(grid_cols):
                    dpg.add_table_column(init_width_or_weight=self.thumbnail_size[0] + 10, width_fixed=True)

                # Add table rows
                for i in range(int(np.ceil(len(self.map_list) / grid_cols))):
                    with dpg.table_row():
                        for j in range(grid_cols):
                            m_id = i * grid_cols + j
                            # Check if maps are remaining
                            if m_id >= len(self.map_list): break
                            m_name = self.map_list[m_id]

                            try:
                                texture = self.map_textures[m_name]
                            except KeyError:
                                texture = self.generic_texture

                            with dpg.table_cell():
                                # Generate button
                                if m_name == self.map_name:
                                    self.map_buttons[m_name] = dpg.add_image_button(texture, label=m_name, callback=self.callback_selectMap, user_data=m_name, tint_color=gui.COLORS["heading"], enabled=False)
                                    self.map_labels[m_name] = dpg.add_text(m_name, color=gui.COLORS["heading"])
                                else:
                                    self.map_buttons[m_name] = dpg.add_image_button(texture, label=m_name, callback=self.callback_selectMap, user_data=m_name)
                                    if "_old" not in m_name:
                                        self.map_labels[m_name] = dpg.add_text(m_name)
                                    else:
                                        self.map_labels[m_name] = dpg.add_text(m_name, color=gui.COLORS["subtle"])
                                
                                # Check target number
                                if self.map_list_tgtnum[m_id] > 0:
                                    dpg.add_text(str(self.map_list_tgtnum[m_id]) + " targets")
                                else:
                                    dpg.add_text("No targets", color=gui.COLORS["error"])

                                # Check inspected status
                                if (self.cur_dir / (m_name + "_inspected.txt")).exists():
                                    dpg.add_text("Inspected", color=gui.COLORS["heading"])
                                else:
                                    # Show inspect button if not currently loaded
                                    if m_name == self.map_name:
                                        dpg.add_text("Not inspected")
                                    else:
                                        dpg.add_button(tag=f"map_btn_ins_{m_id}", label="Mark inspected", callback=self.markInspected, user_data=m_name)

    def deleteMapTable(self):
        """Deletes thumbnail table."""

        if self.map_table is not None:
            if dpg.does_item_exist(self.map_table):
                dpg.delete_item(self.map_table)
            self.map_table = None

        # Delete stats and button as well
        if dpg.does_item_exist("map_stats"):
            dpg.delete_item("map_stats")
        if dpg.does_item_exist("map_butthumb"):
            dpg.delete_item("map_butthumb")

    def loadThumbnails(self, first=False, override=False):
        """Creates thumbnails of lamella maps or load from file if they already exist."""

        # Allow thumbnail setting to be changed by button
        if override:
            self.use_thumbnails = True
            if dpg.does_item_exist("map_butthumb"):
                dpg.configure_item("map_butthumb", label="Creating thumbnails...", enabled=False)

        if not self.use_thumbnails: 
            self.makeGenericThumbnail()

        prev_tex_num = len(self.map_textures)

        now = time.time()
        for m_name in self.map_list:
            if m_name not in self.map_textures.keys():
                self.makeThumbnail(m_name)
            
        if first and self.map_list:
            log(f"Thumbnails loaded in {round((time.time() - now) / len(self.map_list), 1)} s per map.")

        # Create new table if thumbnails changed
        if len(self.map_textures) != prev_tex_num:
            self.deleteMapTable()
            self.makeMapTable()

        if self.use_thumbnails and dpg.does_item_exist("map_butthumb"):
            dpg.delete_item("map_butthumb")

    def makeThumbnail(self, m_name, map_img=None, override=False):
        """Makes thumbnail from map or loads it if file is available."""

        # Make folder 
        if not (self.cur_dir / "thumbnails").exists():
            (self.cur_dir / "thumbnails").mkdir(parents=True)
        
        # If thumbnail file already exists
        thumbnail_path = self.cur_dir / "thumbnails" / (m_name + ".png")
        make_texture = True
        if thumbnail_path.exists():
            map_img = Image.open(thumbnail_path)
        elif self.use_thumbnails or override:
            # Open full map if not opened yet
            if map_img is None:
                map_img = Image.open(self.cur_dir / (m_name + ".png"))
            # Save thumbnail
            map_img.thumbnail(self.thumbnail_size)
            map_img.save(thumbnail_path)
        else:
            make_texture = False

        if make_texture:
            log(f"Creating thumbnail for {m_name}...")
            self.map_textures[m_name] = self.makeTexture(map_img)

    def makeGenericThumbnail(self):
        """Makes a generic schematic lamella thumbnail."""

        size = np.flip(self.thumbnail_size)

        thumbnail = np.zeros(size)
        thumbnail[size[0] // 3: size[0] // 3 * 2, 0: size[1] // 5] = 1
        thumbnail[size[0] // 3: size[0] // 3 * 2, size[1] // 5 * 4: size[1]] = 1
        thumbnail[size[0] // 3: size[0] // 3 * 2, size[1] // 5: size[1] // 5 * 4] = 0.6
        thumbnail[size[0] // 3: size[0] // 3 * 2, size[1] // 5: size[1] // 7 * 2] = 0.3

        # Expansion joints
        thumbnail[size[0] // 5: size[0] // 4, size[1] // 7: size[1] // 7 * 6] = 1
        thumbnail[size[0] // 4 * 3: size[0] // 5 * 4, size[1] // 7: size[1] // 7 * 6] = 1

        self.generic_texture = self.makeTexture(thumbnail)

    def makeTexture(self, image):
        """Creates texture from thumbnail."""

        # Make sure image is of type float
        image = np.array(image)
        if np.max(image) > 1:
            image = image / 255

        # Pad image to size
        map_thumb = np.zeros(np.flip(self.thumbnail_size), dtype=float)
        map_thumb[(self.thumbnail_size[1] - image.shape[0]) // 2: (self.thumbnail_size[1] + image.shape[0]) // 2, (self.thumbnail_size[0] - image.shape[1]) // 2: (self.thumbnail_size[0] + image.shape[1]) // 2] = image

        # Add texture
        map_thumb = np.ravel(np.dstack([map_thumb, map_thumb, map_thumb, np.ones(map_thumb.shape)]))
        return dpg.add_static_texture(width=self.thumbnail_size[0], height=self.thumbnail_size[1], default_value=map_thumb, parent=self.texture_registry)
    
    def show(self):
        dpg.show_item(self.map_window)

    def hide(self):
        dpg.hide_item(self.map_window)

    def update(self, map_name, map_list, map_list_tgtnum):
        """Updates map data and reloads table."""

        self.map_name = map_name
        self.map_list = map_list
        self.map_list_tgtnum = map_list_tgtnum

        self.loadThumbnails()
        self.deleteMapTable()
        self.makeMapTable()

    def updateMap(self, map_name):
        """Updates only map name for highlighting."""

        if map_name in self.map_list:
            if self.map_name in self.map_list:
                # Unhighlight old selected map
                dpg.configure_item(self.map_buttons[self.map_name], tint_color=(255, 255, 255, 255), enabled=True)
                dpg.configure_item(self.map_labels[self.map_name], color=(255, 255, 255, 255))

            # Highlight new selected map
            self.map_name = map_name
            dpg.configure_item(self.map_buttons[self.map_name], tint_color=gui.COLORS["heading"], enabled=False)
            dpg.configure_item(self.map_labels[self.map_name], color=gui.COLORS["heading"])

    def markInspected(self, sender, app_data, user_data):
        """Makes single map as inspected."""

        map_name = user_data
        (self.cur_dir / (map_name + "_inspected.txt")).touch()
        self.deleteMapTable()
        self.makeMapTable()


#!Python
# ===================================================================
#ScriptName SPACEtomo_postAction
# Purpose:      Post action script to setup targets from SPACEtomo runs.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# ===================================================================

import serialem as sem
from SPACEtomo.run_postaction import main

main()



#!/usr/bin/env python
# ===================================================================
# ScriptName:   SPACEtomo_config
# Purpose:      Configurations for deep learning models used in SPACEtomo.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/10/04
# Revision:     v1.2
# Last Change:  2024/09/26: outsourcing of more settings from SPACEtomo_run scripts
#               2024/08/20: added SerialEM connection setting
#               2024/05/02: changed model colors to RGBA
#               2024/03/12: removed max_runs
#               2023/10/04: outsourcing of settings from main SPACEtomo script
# ===================================================================

# SerialEM specific settings
SERIALEM_IP = "127.0.0.1"
SERIALEM_PORT = 48888
SERIALEM_PYTHON_PATH = "C:/Program Files/SerialEM/PythonModules"

DUMMY = False
DEBUG = False

# Acquisition settings
WG_montage_overlap      = 0.15                  # overlap between pieces of whole grid montage
WG_detection_threshold  = 0.1                   # confidence threshold for lamella acceptance

MM_montage_overlap      = 0.25                  # overlap between pieces of medium mag montage
MM_padding_factor       = 1.5                   # padding of estimated lamella size for medium mag montage

aperture_control        = True                  # set to True if SerialEM can control apertures (https://bio3d.colorado.edu/SerialEM/hlp/html/setting_up_serialem.htm#apertures)
objective_aperture      = 70                    # diameter of objective aperture

# Target selection settings
max_iterations          = 10                    # maximum number of iterations for target placement optimization

# Model specific settings (depend on how the model was trained)

# WG model (YOLOv8)
WG_model_file = '2024_07_26_lamella_detect_400nm_yolo8.pt'
WG_model_pix_size = 400.0 # nm/px
WG_model_sidelen = 1024 # px
WG_model_categories = ["broken",           "contaminated",     "good",             "thick",            "wedge",             "gone"          ] 
WG_model_gui_colors = [(255, 125, 0, 255), (255, 215, 0, 255), (50, 150, 50, 255), (59, 92, 128, 255), (87, 138, 191, 255), (200, 0, 0, 255)]
WG_model_nav_colors = [0,                  3,                  1,                  2,                  2,                   0               ]

# Category name     Description                                                         Color
# ----------------- ------------------------------------------------------------------- ----------
# good:             intact lamellae with good thickness                                 green (1)
# thick:            dark lamellae                                                       blue (2)
# wedge/unfinished: lamellae that are good on one end but thick on other end            blue (2)
# contaminated:     lamellae with visible ice blobs                                     yellow (3)
# broken:           lamella with crack, only parts remaining, only one side attached    orange (0)
# gone:             trench without lamella                                              red (0)

# MM model (nnUNet)
MM_model_script = "run_nnUNet.py"
MM_model_folder = 'model'
MM_model_folds = [0, 1, 2, 3, 4]
MM_model_pix_size = 22.83 / 10  # nm/px

#######################################
# Self-check
from pathlib import Path

if Path(WG_model_file).exists():
    NO_WG_MODEL = False
else:
    NO_WG_MODEL = True

if Path(MM_model_folder).exists():
    NO_MM_MODEL = False
else:
    NO_MM_MODEL = True



#!/usr/bin/env python
# ===================================================================
# ScriptName:   SPACEtomo_functions_gui
# Purpose:      Functions necessary to run SPACEtomo and that can be run externally without SerialEM.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/04/26
# Revision:     v1.2
# Last Change:  2024/11/20: added a variety of icons, removed static tag from info box
#               2024/08/09: moved most classes to their own module
#               2024/08/08: added coordinate conversion in MMap class
#               2024/07/30: added combo type to Menu
#               2024/06/05: added Menu class
#               2024/06/04: added status box
#               2024/06/03: added contingency for mrc files containing multiple montages, added info box
#               2024/05/31: added status line, fixed tiles bounds
#               2024/05/30: added rescaling, fixed tile_bounds
#               2024/05/29: outsourced plot clearing
#               2024/05/23: added labels to plot boxes, added padding to edge tiles, added loading of mrc files, fixed tile bounds
#               2024/05/10: added map class
#               2024/04/26
# ===================================================================

import numpy as np
from skimage import transform, draw
import dearpygui.dearpygui as dpg

### GUI CONFIG ###

COLORS = {
    "heading":  (255, 200, 0, 255),
    "error":    (200, 0, 0, 255),
    "subtle":   (255, 255, 255, 64)
}
THEME_COLORS = (
    (87, 138, 191, 255),    # blue
    (229, 242, 255, 255),   # light blue
    (59, 92, 128, 255),     # dark blue
    (200, 0, 0, 255),       # red
    (238, 136, 68, 255)     # orange
)

def configureGlobalTheme():
    """Sets global theme settings for all GUIs."""
    
    # Global theme
    with dpg.theme(tag="global_theme"):
        with dpg.theme_component(dpg.mvButton):
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, 5, category=dpg.mvThemeCat_Core)
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 5, 5)
        with dpg.theme_component(dpg.mvImageButton):
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, 5, category=dpg.mvThemeCat_Core)
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 5, 5)

    # Theme for large buttons
    with dpg.theme(tag="large_btn_theme"):
        with dpg.theme_component(dpg.mvButton):
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 10, 10)
            dpg.add_theme_style(dpg.mvStyleVar_FrameBorderSize, 1)
            dpg.add_theme_color(dpg.mvThemeCol_Text, COLORS["heading"])

    # Theme for small buttons
    with dpg.theme(tag="small_btn_theme"):
        with dpg.theme_component(dpg.mvButton):
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 1, 2)
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, 5)
        with dpg.theme_component(dpg.mvImageButton):
            dpg.add_theme_style(dpg.mvStyleVar_FramePadding, 2, 2)
            dpg.add_theme_style(dpg.mvStyleVar_FrameRounding, 5)

### END CONFIG ###

def cancel_callback():
    pass

def makeLogo(radius=100, stroke=3, oversampling=2):
    radius = radius * oversampling
    stroke = stroke * oversampling

    img = np.zeros([2 * radius, 2 * radius, 4])
    i, j = draw.disk((radius, radius), radius=radius, shape=img.shape[:2])
    img[i, j] = [0, 0, 0, 1]

    i, j = draw.disk((radius, radius), radius=radius - stroke, shape=img.shape[:2])
    gradient = np.array([(x - radius) / radius * np.array([0.965, 0.659, 0.098, 1])  + (1 - (x - radius) / radius) * np.array([0.788, 0.184, 0.157, 1]) for x in i])
    img[i, j] = gradient

    i, j = draw.disk((radius, radius + stroke - 1), radius=radius - stroke * 2, shape=img.shape[:2])
    img[i, j] = [0, 0, 0, 1]

    i, j = draw.disk((radius, radius + stroke - 1), radius=radius - stroke * 3, shape=img.shape[:2])
    img[i, j, :] = 0 

    img = img[radius:2 * radius, :, :]

    img = transform.rescale(img, (1/oversampling, 1/oversampling, 1), anti_aliasing=True)

    logo = np.ravel(img[::-1])
    logo_dims = img.shape

    with dpg.texture_registry():
        dpg.add_static_texture(width=logo_dims[1], height=logo_dims[0], default_value=logo, tag="logo")

    return logo_dims

def makeIconRotation(ccw=False):
    """Makes clockwise arrow icon."""

    if ccw:
        tag = "icon_ccw"
    else:
        tag = "icon_cw"

    if dpg.does_item_exist(tag):
        return tag

    img = np.zeros((13, 13))
    dims = img.shape[:2]

    radius = 5
    i, j, val = draw.circle_perimeter_aa(dims[0] // 2, dims[1] // 2, radius)
    img[i, j] = val

    img[5:10, 9:12] = 0
    img[4, 8:12] = 1
    img[1:4, 11] = 1

    if ccw:
        img = np.flip(img, axis=1)

    tex_img = np.ravel(np.dstack([np.ones(dims), np.ones(dims), np.ones(dims), img]))

    with dpg.texture_registry():
        dpg.add_static_texture(width=dims[1], height=dims[0], default_value=tex_img, tag=tag)

    return tag

def makeIconFlip(horizontal=False):
    """Makes double sided arrow icon."""

    if horizontal:
        tag = "icon_fliph"
    else:
        tag = "icon_flipv"

    if dpg.does_item_exist(tag):
        return tag

    img = np.zeros((13, 13))
    dims = img.shape[:2]

    i, j, val = draw.line_aa(1, dims[0] // 2, 3, dims[0] // 2 + 2)
    img[i, j] = val
    i, j, val = draw.line_aa(1, dims[0] // 2, 3, dims[0] // 2 - 2)
    img[i, j] = val

    i, j, val = draw.line_aa(dims[1] - 2, dims[0] // 2, dims[1] - 4, dims[0] // 2 + 2)
    img[i, j] = val
    i, j, val = draw.line_aa(dims[1] - 2, dims[0] // 2, dims[1] - 4, dims[0] // 2 - 2)
    img[i, j] = val

    i, j, val = draw.line_aa(1, dims[0] // 2, dims[1] - 2, dims[0] // 2)
    img[i, j] = val

    if horizontal:
        img = img.T

    tex_img = np.ravel(np.dstack([np.ones(dims), np.ones(dims), np.ones(dims), img]))

    with dpg.texture_registry():
        dpg.add_static_texture(width=dims[1], height=dims[0], default_value=tex_img, tag=tag)

    return tag

def makeIconDelete():
    """Makes trashcan icon."""

    tag = "icon_delete"

    if dpg.does_item_exist(tag):
        return tag

    img = np.zeros((13, 13))
    dims = img.shape[:2]
    radius = 2

    # Corners
    i, j, val = draw.circle_perimeter_aa(9, 4, radius)
    img[i, j] = val
    i, j, val = draw.circle_perimeter_aa(9, 8, radius)
    img[i, j] = val
    # Clean center
    img[0:10, 3:10] = 0
    img[0:12, 5:8] = 0
    # Bottom
    i, j, val = draw.line_aa(11, 4, 11, 8)
    img[i, j] = val
    # Sides
    i, j, val = draw.line_aa(4, 2, 9, 2)
    img[i, j] = val
    i, j, val = draw.line_aa(5, 4, 9, 4)
    img[i, j] = val
    i, j, val = draw.line_aa(5, 6, 9, 6)
    img[i, j] = val
    i, j, val = draw.line_aa(5, 8, 9, 8)
    img[i, j] = val
    i, j, val = draw.line_aa(4, 10, 9, 10)
    img[i, j] = val
    # Top

    i, j, val = draw.line_aa(3, 3, 1, 5)
    img[i, j] = val
    i, j, val = draw.line_aa(1, 7, 3, 9)
    img[i, j] = val
    i, j, val = draw.line_aa(3, 1, 3, 11)
    img[i, j] = val
    img[1, 6] = 1

    tex_img = np.ravel(np.dstack([np.ones(dims), np.ones(dims), np.ones(dims), img]))

    with dpg.texture_registry():
        dpg.add_static_texture(width=dims[1], height=dims[0], default_value=tex_img, tag=tag)

    return tag

def makeIconEdit():
    """Makes pencil icon."""

    tag = "icon_edit"

    if dpg.does_item_exist(tag):
        return tag
    img = np.zeros((13, 13))
    dims = img.shape[:2]

    # Pencil sides
    i, j, val = draw.line_aa(7, 3, 0, 10)
    img[i, j] = val
    i, j, val = draw.line_aa(9, 5, 2, 12)
    img[i, j] = val

    # Pencils top/bottom
    i, j, val = draw.line_aa(0, 10, 2, 12)
    img[i, j] = val
    i, j, val = draw.line_aa(7, 3, 9, 5)
    img[i, j] = val
    i, j, val = draw.line_aa(2, 8, 4, 10)
    img[i, j] = val

    # Pencil tip
    i, j, val = draw.line_aa(7, 3, 10, 2)
    img[i, j] += val
    i, j, val = draw.line_aa(9, 5, 10, 2)
    img[i, j] += val

    # Paper
    i, j, val = draw.line_aa(2, 0, 12, 0)
    img[i, j] = val
    i, j, val = draw.line_aa(2, 0, 2, 5)
    img[i, j] = val
    i, j, val = draw.line_aa(12, 0, 12, 8)
    img[i, j] = val
    i, j, val = draw.line_aa(9, 8, 12, 8)
    img[i, j] = val

    img = np.clip(img, 0, 1)

    tex_img = np.ravel(np.dstack([np.ones(dims), np.ones(dims), np.ones(dims), img]))

    with dpg.texture_registry():
        dpg.add_static_texture(width=dims[1], height=dims[0], default_value=tex_img, tag=tag)

    return tag

def window_size_change(logo_dims):
    # Update items anchored to side of window
    dpg.set_item_pos("logo_img", pos=(10, dpg.get_viewport_client_height() - 40 - logo_dims[0]))
    dpg.set_item_pos("logo_text", pos=(10 + logo_dims[1] / 2 - (30), dpg.get_viewport_client_height() - 40 - logo_dims[0] / 2))
    dpg.set_item_pos("version_text", pos=(10 + logo_dims[1] / 2 - (30), dpg.get_viewport_client_height() - 27 - logo_dims[0] / 2))

def askForSave():
    # TODO check for saving conditions
    dpg.stop_dearpygui()

def fileNav(tag, callback, dir=False, extensions=[]):
    with dpg.file_dialog(directory_selector=dir, show=False, callback=callback, tag=tag, cancel_callback=cancel_callback, width=700 ,height=400): 
        dpg.add_file_extension(".*") 
        for ext in extensions:
            dpg.add_file_extension(ext, color=COLORS["heading"])


class StatusLine:
    def __init__(self, item=None) -> None:
        self.status = ""
        if item is None:
            self.item = dpg.add_text(default_value=self.status)
        else:
            self.item = item
        dpg.hide_item(self.item)
        dpg.configure_item(self.item, color=COLORS["subtle"])
        self.processing_box = None

    def update(self, status="", color=None, box=False):
        self.status = status

        if dpg.does_item_exist(self.processing_box):
            dpg.delete_item(self.processing_box)
            dpg.split_frame()

        dpg.set_value(self.item, self.status)
        if color is not None:
            dpg.configure_item(self.item, color=color)
        else:
            dpg.configure_item(self.item, color=(255, 255, 255, 255))
        if self.status == "":
            dpg.hide_item(self.item)
        else:
            dpg.show_item(self.item)
            if box:
                self.processing_box = showInfoBox("PROCESSING", self.status, options=None)

def showInfoBox(title, message, callback=None, options=[], options_data=[]):
    """Shows info or confirmation pop up box."""

    # Set color
    if title == "ERROR":
        color = COLORS["error"]
    else:
        color = (255, 255, 255, 255)

    viewport_size = np.array((dpg.get_viewport_client_width(), dpg.get_viewport_client_height()))

    # Create popup window
    with dpg.window(label=title, modal=True, no_close=True) as infobox:
        dpg.add_text(message, color=color)
        if options is not None:
            if callback is None:
                callback = lambda: dpg.delete_item(infobox)
            dpg.add_text()
            with dpg.group(horizontal=True) as infobtns:
                if len(options) > 0:
                    for o, option in enumerate(options):
                        if o < len(options_data):
                            user_data = options_data[o]
                        else:
                            user_data = o
                        dpg.add_button(label=option, user_data=[infobox, user_data], callback=callback)
                else:
                    dpg.add_button(label="OK", callback=callback)

    # Wait for next frame so size and position can be adjusted
    dpg.split_frame()

    window_size = np.array((dpg.get_item_width(infobox), dpg.get_item_height(infobox)))
    dpg.set_item_pos(infobox, pos=(viewport_size - window_size) / 2)    # pos needs to be a float from dpg>=2.0

    # Center buttons
    if options is not None:
        group_size=np.array(dpg.get_item_rect_size(infobtns))
        dpg.set_item_pos(infobtns, ((window_size[0] - group_size[0]) / 2, dpg.get_item_pos(infobtns)[1]))    # pos needs to be a float from dpg>=2.0

    return infobox



#!/usr/bin/env python
# ===================================================================
# ScriptName:   GUI
# Purpose:      User interface for inspecting and labeling lamella grid atlases collected with SPACEtomo.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/04/26
# Revision:     v1.2
# Last Change:  2024/07/18: reworked GUI to class
# #             2024/06/07: added lamella table row highlighting, added save coords in YOLO format
#               2024/06/05: introduced Menu class, added shortcuts for most buttons
#               2024/06/04: added pix size check for export, added rescaling of mrc file
#               2024/06/03: added include empty tiles option, added save json option
#               2024/05/31: added status line, fixed tiles bounds, added next button, added hide/showElements
#               2024/05/30: added rescaling, saving as png, fixed export tiles
#               2024/05/29: outsourced plot clearing
#               2024/05/23: added lamella labels on plot, added padding when exporting edge tiles, fixed plotTiles
#               2024/05/10: added reading and writing of YOLO format labels, added saving of tiles with labels, added map class
#               2024/05/02: added lamella color from config, added reset zoom, delegated plot generation, call lamella detection
#               2024/04/30: added lamella focus buttons, outline scaling, list sorting
#               2024/04/26: Copy most of SPACEtomo_tgt
# ===================================================================

import sys
from SPACEtomo.modules.utils import log
try:
    import dearpygui.dearpygui as dpg
except:
    log("ERROR: DearPyGUI module not installed! If you cannot install it, please run the target selection GUI from an external machine.")
    sys.exit()
import argparse
from pathlib import Path
from SPACEtomo.modules.gui.lam_sel import LamellaGUI
from SPACEtomo.modules.gui.tgt_sel import TargetGUI

def run_process(gui, path="", auto_close=False):
    dpg.create_context()
    main = gui(path, auto_close=auto_close)
    main.show()
    dpg.destroy_context()
    
def main():
    # Process arguments
    parser = argparse.ArgumentParser(description="Calls GUI to inspect whole grid maps and detected lamellae.")
    parser.add_argument("gui", type=str, default="targets", help="Name of GUI to be opened. Options: targets, lamella")
    parser.add_argument("map_file", nargs="?", type=str, default="", help="Absolute path to a map [.png].")
    parser.add_argument("--auto_close", action="store_true", help="Auto-close GUI after inspection.")
    args = parser.parse_args()

    # Get GUI
    if args.gui == "targets":
        gui = TargetGUI
    elif args.gui == "lamella":
        gui = LamellaGUI
    else:
        raise ValueError(f"No GUI found with name {args.gui}!")

    # Check map file
    map_file = ""
    if args.map_file != "": 
        map_file = Path(args.map_file)
        if not map_file.exists():
            map_file = ""
            log("WARNING: Map file does not exist! Attempting to open GUI without map...")

    # Run GUI
    run_process(gui, map_file, args.auto_close)
    

if __name__ == "__main__":
    main()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   SPACEtomo_monitor
# Purpose:      Monitors an external directory and runs a target selection deep learning model on medium mag montages of lamella and generates a segmentation that can be used for PACEtomo target selection.
#               More information at http://github.com/eisfabian/PACEtomo
# Usage:        python SPACEtomo_monitor.py [--dir MAP_DIR] [--gpu 0,1,2,3] [--plot]
# Author:       Fabian Eisenstein
# Created:      2023/10/05
# Revision:     v1.2
# Last Change:  2024/08/20: adjusted to refactor
#               2024/04/09: deleted run file upon start to avoid problems upon restart
#               2024/04/04: added call to tgt GUI
#               2024/03/25: fixes after Krios 3 test
#               2024/03/12: replaced script settings with arguments, added stepwise inclusion of params as they appear
#               2024/03/11: simplification and fixes
#               2024/03/07: rewrote updateQueue function, changed runs and queue to single json file, moved functions to functions_ext
#               2024/02/14: added WG map monitoring and lamella detection, set timer step down to 1 s, added multi GPU usage
#               2023/12/04: added check for memory error
#               2023/11/15: added file being added to queue before submitting run to give time to finish copying
#               2023/10/31: removed wait for settings
#               2023/10/18: fixes after local test
#               2023/10/05: 
# ===================================================================

import sys
import time
import argparse
import subprocess
from pathlib import Path
import torch

from SPACEtomo.modules.mod_wg import WGModel
from SPACEtomo.modules.scope import ImagingParams
import SPACEtomo.modules.ext as space_ext
from SPACEtomo.modules.utils import log
from SPACEtomo import __version__

def main():
    # Process arguments
    parser = argparse.ArgumentParser(description='Monitors an external directory and runs lamella detection, lamella segmentation and target selection on appropiate files.')
    parser.add_argument('--dir', dest='map_dir', type=str, default=None, help='Absolute path to folder to be monitored. This should be the same directory that was set in SPACEtomo on the SerialEM PC. (Default: Folder this script is run from)')
    parser.add_argument('--gpu', dest='gpu', type=str, default="0", help='Comma-separated IDs of GPUs to use. (Default: 0)')
    parser.add_argument('--plot', dest='save_plot', action='store_true', help='Create plots of intermediate target selection steps (slow).')
    args = parser.parse_args()

    if args.map_dir is not None: 
        if Path(args.map_dir).exists():
            MAP_DIR = Path(args.map_dir)
        else:
            log("ERROR: Folder does not exist!")
            sys.exit()
    else:
        MAP_DIR = Path.cwd()

    gpus = args.gpu.split(",")
    gpu_list = [int(gpu) for gpu in gpus]
    gpu_list = gpu_list if torch.cuda.is_available() else []        # Check GPU availability

    save_plot = args.save_plot

    # Load model
    WG_model = WGModel()

    MM_model = None
    mic_params = None
    tgt_params = None

    # Move previous run file
    if (MAP_DIR / "SPACE_runs.json").exists():
        (MAP_DIR / "SPACE_runs.json").replace(MAP_DIR / "SPACE_runs.json~")

    # Indicator for monitors running (lamella detection, lamella segmentation, target selection)
    status = "(oxx)"

    # Set start time
    start_time = time.time()
    next_time = start_time + 60

    log(f"SPACEtomo Version {__version__}\n")
    log(f"Start monitoring {MAP_DIR} for all maps... {status}")
    log(f"Using GPUs: {gpu_list}\n")

    while True:
        now = time.time()
        if now > next_time:
            next_time = now + 60
            log(f"##### Running for {int(round((now - start_time) / 60))} min... {status} #####")
        
        # Check if mic_params exist
        if mic_params is None and (MAP_DIR / "mic_params.json").exists():
            status = "(oox)"
            log(f"NOTE: Microscope parameters found. Including lamella segmentation. {status}")
            # Instantiate mic params from settings
            mic_params = ImagingParams(None, None, None, file_dir=MAP_DIR)        
            # Load model
            MM_model = space_ext.MMModel()

        # Check if tgt_params exist
        if tgt_params is None and (MAP_DIR / "tgt_params.json").exists():
            status = "(ooo)"
            log(f"NOTE: Target parameters found. Including target setup. {status}")
            # Reimport mic params to get Rec params
            mic_params = space_ext.MicParams_ext(MAP_DIR) 
            MM_model.setDimensions(mic_params)
            # Instantiate tgt params from settings
            tgt_params = space_ext.TgtParams(file_dir=MAP_DIR, MM_model=MM_model)

            log("Opening target selection GUI for inspection...")
            subprocess.Popen([sys.executable, Path(__file__).parent / "GUI.py", "targets", MAP_DIR])

        # Run queue
        space_ext.updateQueue(MAP_DIR, WG_model, MM_model, mic_params, tgt_params, gpu_list, save_plot)
        time.sleep(5)

if __name__ == "__main__":
    main()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   gui_plt
# Purpose:      User interface utility to generate plot
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/08
# Revision:     v1.2
# Last Change:  2024/11/18: moved all texture deletions after plot deletions
#               2024/11/12: added annotations
#               2024/11/08: added border dragging of boxes
#               2024/08/08: separated from gui.py 
# ===================================================================

import time
import numpy as np
import dearpygui.dearpygui as dpg

from SPACEtomo.modules.gui.gui import COLORS

class Plot:
    """Class to hold and organize all plotted objects."""

    def __init__(self, tag) -> None:
        self.plot = None
        self.x_axis = None
        self.y_axis = None

        self.tag = tag

        self.img = []
        self.boxes = []
        self.overlays = []
        self.series = []
        self.drag_points = []
        self.annotations = []

        self.bounds = np.zeros((2, 2))  # [[xmin, xmax], [ymin, ymax]]

    def sortList(self, list_name, sortable):
        """Sorts list of plots according to label."""

        unsorted_list = getattr(self, list_name)
        if isinstance(unsorted_list, list) and len(unsorted_list) > 1:
            if hasattr(unsorted_list[0], sortable):
                sortable_list = [getattr(item, sortable) for item in unsorted_list]
                sorted_list = [item for _, item in sorted(zip(sortable_list, unsorted_list))]
                setattr(self, list_name, sorted_list)

    def makePlot(self, x_axis_label="x", y_axis_label="y", **kwargs):
        """Creates base plot object."""

        self.plot = dpg.add_plot(tag=self.tag, **kwargs)

        self.x_axis = dpg.add_plot_axis(dpg.mvXAxis, label=x_axis_label, parent=self.plot)#, tag="x_axis")
        self.y_axis = dpg.add_plot_axis(dpg.mvYAxis, label=y_axis_label, parent=self.plot)#, tag="y_axis")

    def resetZoom(self):
        """Auto zooms to plot bounds."""

        # Fit axes
        dpg.fit_axis_data(self.x_axis)
        dpg.fit_axis_data(self.y_axis)

    def updateLabel(self, label=""):
        """Updates the plot heading."""

        dpg.set_item_label(self.plot, label)

    def addImg(self, texture, bounds, binning=1, label=""):
        """Plots an image."""

        bounds = np.array(bounds)
        self.img.append({"label": label, "tex": texture, "plot": dpg.add_image_series(texture, bounds_min=(bounds[0, 0], bounds[1, 0]), bounds_max=(bounds[0, 1], bounds[1, 1]), parent=self.x_axis), "binning": binning})
        self.updateBounds(bounds)

    def addSeries(self, x_vals, y_vals, label="", theme=""):
        """Plots a scatter series."""

        self.series.append({"label": label, "plot": dpg.add_scatter_series(list(x_vals), list(y_vals), parent=self.x_axis)})
        if theme and dpg.does_item_exist(theme):
            dpg.bind_item_theme(self.series[-1]["plot"], theme)

    def addAnnotation(self, text, x, y, offset=(10, -10), label="", color=COLORS["heading"]):
        """Plots annotation label."""

        self.annotations.append({"label": label, "annotation": dpg.add_plot_annotation(label=text, default_value=(x, y), offset=offset, color=color, clamped=False, parent=self.plot)})

    def addDragPoint(self, x, y, label, callback=None, user_data=None, color=(255, 255, 255, 255)):
        """Creates a drag point."""

        self.drag_points.append({"label": label, "plot": dpg.add_drag_point(default_value=(x, y), user_data=user_data, color=color, label=label, callback=callback, parent=self.plot)})

    def addOverlay(self, texture, bounds, label):
        """Plots an image (meant for small overlays)."""

        self.overlays.append({"label": label, "tex": texture, "plot": dpg.add_image_series(texture, bounds_min=(bounds[0][0], bounds[1][0]), bounds_max=(bounds[0][1], bounds[1][1]), parent=self.x_axis)})

    def addCustomPlot(self, list, label, plot, theme=None):
        """Adds custom series to any list of self."""

        list.append({"label": label, "plot": plot})
        if theme:
            dpg.bind_item_theme(plot, theme)

    def updateBounds(self, bounds):
        """Updates bounds of plot."""

        self.bounds[0, 0] = min(self.bounds[0, 0], bounds[0, 0])        # xmin
        self.bounds[0, 1] = max(self.bounds[0, 1], bounds[0, 1])        # xmax
        self.bounds[1, 0] = min(self.bounds[1, 0], bounds[1, 0])        # ymin
        self.bounds[1, 1] = max(self.bounds[1, 1], bounds[1, 1])        # ymax

    def withinBounds(self, coords):
        """Checks if coordinates are within the plot bounds."""

        x, y = coords
        if self.bounds[0, 0] <= x <= self.bounds[0, 1] and self.bounds[1, 0] <= y <= self.bounds[1, 1]:
            return True
        else:
            return False
        
    def updateImg(self, id, texture=None, bounds=None):
        """Updates texture of plotted image series."""

        if texture is not None:
            # Old texture
            old_tex = self.img[id]["tex"]

            # Update texture
            self.img[id]["tex"] = texture
            dpg.configure_item(self.img[id]["plot"], texture_tag=texture)

            # Delete old texture
            if dpg.does_item_exist(old_tex): 
                dpg.delete_item(old_tex)
                dpg.split_frame(delay=10)                       # helps to reduce Segmentation fault crashes

        # Update bounds
        if bounds is not None:
            dpg.configure_item(self.img[id]["plot"], bounds_min=(bounds[0][0], bounds[1][0]), bounds_max=(bounds[0][1], bounds[1][1]))

    def updateOverlay(self, id, texture=None, bounds=None):
        """Updates texture of plotted overlay image series."""

        if texture is not None:
            # Old texture
            old_tex = self.overlays[id]["tex"]

            # Update texture
            self.overlays[id]["tex"] = texture
            dpg.configure_item(self.overlays[id]["plot"], texture_tag=texture)

            # Delete old texture
            if dpg.does_item_exist(old_tex): 
                dpg.delete_item(old_tex)
                dpg.split_frame(delay=10)                       # helps to reduce Segmentation fault crashes

        # Update bounds
        if bounds is not None:
            dpg.configure_item(self.overlays[id]["plot"], bounds_min=(bounds[0][0], bounds[1][0]), bounds_max=(bounds[0][1], bounds[1][1]))

    """Methods to find specific plot elements whose label contains a keyword."""
    def getImgByKeyword(self, keyword):
        return [item["label"] for item in self.img if keyword in item["label"]]
    
    def getSeriesByKeyword(self, keyword):
        return [item["label"] for item in self.series if keyword in item["label"]]
    
    def getDragPointsByKeyword(self, keyword):
        return [item["label"] for item in self.drag_points if keyword in item["label"]]

    def getOverlaysByKeyword(self, keyword):
        return [item["label"] for item in self.overlays if keyword in item["label"]]    

    def clearImg(self, labels=[], delete_textures=True):
        """Delete all or some plotted images and their textures."""

        remaining_list = []
        for img in self.img:
            if labels and img["label"] not in labels:       # Option for removing only certain images
                remaining_list.append(img)
                continue
            if dpg.does_item_exist(img["plot"]): 
                dpg.delete_item(img["plot"])
            if delete_textures and dpg.does_item_exist(img["tex"]): 
                dpg.delete_item(img["tex"])
                dpg.split_frame(delay=10)                   # helps to reduce Segmentation fault crashes

        self.img = remaining_list

    def clearSeries(self, labels=[]):
        """Delete all or some plotted scatter series."""

        remaining_list = []
        for series in self.series:
            if labels and series["label"] not in labels:    # Option for removing only certain series
                remaining_list.append(series)
                continue
            if dpg.does_item_exist(series["plot"]): 
                dpg.delete_item(series["plot"])
        self.series = remaining_list

    def clearBoxes(self):
        """Delete all plotted boxes."""

        for box in self.boxes:
            box.remove()
            if dpg.does_item_exist(box):
                dpg.delete_item(box)
        self.boxes = []

    def clearOverlays(self, labels=[], delete_textures=True):
        """Delete all or some plotted overlays and their textures."""

        remaining_list = []
        for overlay in self.overlays:
            if labels and overlay["label"] not in labels:   # Option for removing only certain series
                remaining_list.append(overlay)
                continue
            if dpg.does_item_exist(overlay["plot"]): 
                dpg.delete_item(overlay["plot"])
            if delete_textures and dpg.does_item_exist(overlay["tex"]):
                dpg.delete_item(overlay["tex"])
                dpg.split_frame(delay=10)                   # helps to reduce Segmentation fault crashes

        self.overlays = remaining_list

    def clearDragPoints(self, labels=[]):
        """Delete all or some drag points."""

        remaining_list = []
        for point in self.drag_points:
            if labels and point["label"] not in labels:     # Option for removing only certain series
                remaining_list.append(point)
                continue
            if dpg.does_item_exist(point["plot"]):
                dpg.delete_item(point["plot"])
        self.drag_points = remaining_list

    def clearAnnotations(self, labels=[]):
        """Delete all or some annotations."""

        remaining_list = []
        for annotation in self.annotations:
            if labels and annotation["label"] not in labels:     # Option for removing only certain series
                remaining_list.append(annotation)
                continue
            if dpg.does_item_exist(annotation["annotation"]):
                dpg.delete_item(annotation["annotation"])
        self.annotations = remaining_list

    def clearAll(self):
        """Reset the plot."""

        self.clearImg()
        self.clearSeries()
        self.clearBoxes()
        self.clearOverlays()
        self.clearDragPoints()
        self.clearAnnotations()
        self.bounds = np.zeros((2, 2))


class PlotBox:
    """Class for bounding boxes on plot."""
    
    def __init__(self, coords, parent, color=COLORS["error"], thickness=1) -> None:
        """Initializes box with bounds at same coords defined by 2 points."""

        self.p1 = coords
        self.p2 = coords

        self.parent = parent
        self.color = color
        self.thickness = 1
        self.rect = None
        self.annotation = None

        # Keep track of update mode
        self.update_mode = None
        self.update_start = None

    @property
    def center(self):
        return (self.p1 + self.p2) / 2
    
    @property
    def top(self):
        return max([self.p1[1], self.p2[1]])
    
    @property
    def bottom(self):
        return min([self.p1[1], self.p2[1]])
    
    @property
    def height(self):
        return self.top - self.bottom
    
    @property
    def left(self):
        return min([self.p1[0], self.p2[0]])
    
    @property
    def right(self):
        return max([self.p1[0], self.p2[0]])
    
    @property
    def width(self):
        return self.right - self.left
    
    def startUpdate(self, coords):
        """Determines which update mode is appropriate depending on where coords are."""

        self.update_start = coords
        margin = 0.1 # 10 %

        # Top edge
        if self.top - margin * self.height <= coords[1] <= self.top + margin * self.height: 
            # Left
            if self.left - margin * self.width <= coords[0] <= self.left + margin * self.width:
                self.update_mode = ["top", "left"]
            # Right
            elif self.right - margin * self.width <= coords[0] <= self.right + margin * self.width:
                self.update_mode = ["top", "right"]
            # Edge
            else:
                self.update_mode = ["top"]
        # Bottom edge
        elif self.bottom - margin * self.height <= coords[1] <= self.bottom + margin * self.height: 
            # Left
            if self.left - margin * self.width <= coords[0] <= self.left + margin * self.width:
                self.update_mode = ["bottom", "left"]
            # Right
            elif self.right - margin * self.width <= coords[0] <= self.right + margin * self.width:
                self.update_mode = ["bottom", "right"]
            # Edge
            else:
                self.update_mode = ["bottom"]
        # Left edge
        elif self.left - margin * self.width <= coords[0] <= self.left + margin * self.width:
            self.update_mode = ["left"]
        # Right edge
        elif self.right - margin * self.width <= coords[0] <= self.right + margin * self.width:
            self.update_mode = ["right"]
        # Edge
        else:
            self.update_mode = ["top", "bottom", "left", "right"]

    def endUpdate(self):
        """Finished update mode."""

        self.update_mode = None
        self.update_start = None

        # Make sure p1 is top left and p2 is bottom right
        top_left = np.array([self.left, self.top])
        bottom_right = np.array([self.right, self.bottom])
        self.p1, self.p2 = top_left, bottom_right

    def update(self, coords):
        """Updates dependent on mode."""

        shift = coords - self.update_start

        if "top" in self.update_mode:
            self.p1[1] = self.p1[1] + shift[1]
        if "bottom" in self.update_mode:
            self.p2[1] = self.p2[1] + shift[1]
        if "left" in self.update_mode:
            self.p1[0] = self.p1[0] + shift[0]
        if "right" in self.update_mode:
            self.p2[0] = self.p2[0] + shift[0]

        dpg.configure_item(self.rect, pmin=self.p1, pmax=self.p2)

        if dpg.does_item_exist(self.annotation):
            dpg.set_value(self.annotation, (self.right, self.top))

        self.update_start = coords

    def updateP2(self, coords):
        """Updates second point to define bounds."""

        self.p2 = coords
        if self.rect is None:
            self.draw()
        else:
            dpg.configure_item(self.rect, pmax=self.p2)

        if dpg.does_item_exist(self.annotation):
            dpg.set_value(self.annotation, (self.right, self.top))

    def updateP1(self, coords):
        """Updates first point to redefine bounds."""

        self.p1 = coords
        self.p2 = coords
        dpg.configure_item(self.rect, pmin=self.p1, pmax=self.p2)

        if dpg.does_item_exist(self.annotation):
            dpg.set_value(self.annotation, (self.right, self.top))

    def updateThickness(self, thickness):
        """Updates border thickness of box plot."""

        self.thickness = thickness
        if dpg.does_item_exist(self.rect):
            dpg.configure_item(self.rect, thickness=self.thickness)

    def updateColor(self, color):
        """Updates border color of box plot."""

        self.color = color
        if dpg.does_item_exist(self.rect):
            dpg.configure_item(self.rect, color=self.color)
        if dpg.does_item_exist(self.annotation):
            dpg.configure_item(self.annotation, color=self.color)

    def within(self, coords, margin=0):
        """Checks if coords are within the box (or within margin of the box)."""

        if min(self.p1[0], self.p2[0]) - margin <= coords[0] < max(self.p1[0], self.p2[0]) + margin and min(self.p1[1], self.p2[1]) - margin <= coords[1] < max(self.p1[1], self.p2[1]) + margin:
            return True
        else:
            return False
        
    def drawLabel(self, label, offset=(10, -10)):
        """Draws an annotation on the plot."""

        if dpg.does_item_exist(self.annotation):
            dpg.delete_item(self.annotation)
        self.annotation = dpg.add_plot_annotation(label=label, default_value=(self.right, self.top), offset=offset, color=self.color, clamped=False, parent=self.parent)

    def draw(self):
        """Draws the box on the plot."""

        self.rect = dpg.draw_rectangle(pmin=self.p1, pmax=self.p2, color=self.color, thickness=self.thickness, parent=self.parent)

    def remove(self):
        """Deletes the box and annotation from the plot."""
        
        if dpg.does_item_exist(self.annotation):
            dpg.delete_item(self.annotation)
        dpg.delete_item(self.rect)



### WARNING: This setup.py should only be used when installing the minimal environment using Python 3.6 (old versions of pip)

from setuptools import setup, find_packages
import os

# Define the base directory of the package
base_dir = os.path.dirname(os.path.abspath(__file__))

# Read the __version__ from the package's __init__.py file
def get_version():
    version_file = os.path.join(base_dir, 'SPACEtomo', '__init__.py')
    with open(version_file) as f:
        for line in f:
            if line.startswith('__version__'):
                # Extract the version number from the line
                return line.split('=')[-1].strip().strip('"').strip("'")
    raise RuntimeError("Unable to find version string.")

setup(
    name="SPACEtomo",
    version=get_version(), # Use the __version__ from __init__.py
    description="Smart Parallel Automated Cryo Electron tomography (SPACEtomo) is a package that - together with SerialEM - completely automates the cryoET data acquisition workflow.",
    long_description=open('README.md').read(), # Read the long description from README.md
    long_description_content_type="text/markdown",
    python_requires=">=3.6,!=3.7.*,!=3.8.*,!=3.10.*", # SerialEM does not support 3.10
    license=open('LICENSE').read(),
    keywords=["SPACEtomo", "cryoET", "tomography", "electron-tomography", "SerialEM"],
    maintainer="Fabian Eisenstein",
    maintainer_email="fabian.eisenstein@ethz.ch",
    packages=find_packages(), # Automatically find packages
    install_requires=[
        "matplotlib>=3.5; python_version>='3.9'",
        "mrcfile>=1.5.0; python_version>='3.9'",
        "numpy>=1.26.4; python_version>='3.9'",
        "Pillow>=10.4.0; python_version>='3.9'",
        "scipy>=1.12.0; python_version>='3.9'",
        "scikit-image>=0.22.0; python_version>='3.9'",

        "matplotlib>=3.3; python_version<'3.7'",
        "mrcfile>=1.5; python_version<'3.7'",
        "numpy>=1.19; python_version<'3.7'",
        "Pillow>=8.0; python_version<'3.7'",
        "scipy>=1.5; python_version<'3.7'",
        "scikit-image>=0.17; python_version<'3.7'",
        "imageio==2.15.0; python_version<'3.7'", # max version 2.16.0 needs numpy>=1.20
    ],
    extras_require={
        "gui": [
            "dearpygui>=1.10.0; python_version>='3.9'",
            ],
        "predict": [
            "torch>=2.2.0; python_version>='3.9'", 
            "torchvision>=0.17.0; python_version>='3.9'", 
            "nnunetv2>=2.2.1; python_version>='3.9'", 
            "ultralytics>=8.1.12; python_version>='3.9'",
            "ultralytics>=8.1.12,<8.3.10; python_version=='3.9'",
            ],
    },
    entry_points={
        "console_scripts": [
            "SPACEtomo=SPACEtomo.CLI:main",
            "SPACEmonitor=SPACEtomo.run_monitor:main",
            "SPACEmodel=SPACEtomo.setup_model:main",
            "SPACEconfig=SPACEtomo.setup_config:main",
        ]
    },
    url="https://github.com/eisfabian/SPACEtomo",
    project_urls={
        "Homepage": "https://github.com/eisfabian/SPACEtomo",
        "Documentation": "https://github.com/eisfabian/SPACEtomo/blob/main/README.md",
        "Repository": "https://github.com/eisfabian/SPACEtomo",
        "Issues": "https://github.com/eisfabian/SPACEtomo/issues",
    },
)



#!/usr/bin/env python
# ===================================================================
# Purpose:      Functions for utilities needed by other packages and scripts.
# Author:       Fabian Eisenstein
# Created:      2024/07/18
# Last Change:  2024/09/24: updated log to handle debug and line breaks
#               2024/09/04: converted settings to configparser
#               2024/09/02: added dummy_replace
#               2024/08/19: moved and updated writeTargets, moved writeMrc, moved saveSettings
# ===================================================================

import os
import sys
import time
import json
import struct
import numpy as np
import mrcfile
import hashlib
import subprocess
import concurrent.futures
from pathlib import Path
from functools import wraps
from configparser import ConfigParser
from PIL import Image
Image.MAX_IMAGE_PIXELS = None           # removes limit for large images

# Check SerialEM availability
try:
    import serialem as sem
    # Check if PySEMSocket is connected
    try:
        sem.Delay(0, "s")
        SERIALEM = True
    except sem.SEMmoduleError:
        SERIALEM = False
except ModuleNotFoundError:
    SERIALEM = False

from SPACEtomo import __version__
from SPACEtomo import config

# Check if torch and GPU are available
try:
    from torch.cuda import is_available
    if is_available():
        DEVICE = "cuda"
    else:
        DEVICE = "cpu"
except:
    DEVICE = "cpu"

# Write to SerialEM log or print
def log(text, color=0, style=0):
    if text.startswith("DEBUG:") and not config.DEBUG:
        return
    if SERIALEM:
        if text.startswith("NOTE:"):
            color = 4
        elif text.startswith("WARNING:"):
            color = 5
        elif text.startswith("ERROR:"):
            color = 2
            style = 1 
        elif text.startswith("DEBUG:"):
            color = 1

        sem.SetNextLogOutputStyle(style, color)
        sem.EchoBreakLines(text)
    else:
        print(text)

def getCurDir():
    """Gets current directory from SerialEM or from main script."""

    if SERIALEM:
        return Path(sem.ReportDirectory())
    else:
        return Path.cwd()

# Find index in list of dicts for specific key, value pair
def findIndex(lst, key, val):
    if len(lst) > 0:
        if isinstance(lst[0], dict):
            for i, dic in enumerate(lst):
                if dic[key] == val:
                    return i
    return None

# Read generic column based text file
def loadColFile(file, force_type=str):
    data = []
    if os.path.exists(file):
        with open(file, "r") as f:
            lines = f.readlines()
        for line in lines:
            cols = line.strip().split()
            if len(cols) > 0:
                data.append([])
                for col in cols:
                    data[-1].append(force_type(col))
    return data

# Convert numpy arrays to list for json (https://stackoverflow.com/a/65354261)
def convertArray(array):
    if hasattr(array, "tolist"):
        return {"$array": array.tolist()}       # convert to list in dict with tag
    raise TypeError(array)

def revertArray(array):
    if len(array) == 1:                         # check only dicts with one key
        key, value = next(iter(array.items()))
        if key == "$array":                     # if the key is the tag, convert
            return np.array(value)
    return array

def toNumpy(img):
    """Loads image with slightly faster (~3%) conversion from PIL.Image to numpy.array (https://uploadcare.com/blog/fast-import-of-pillow-images-to-numpy-opencv-arrays/)"""
    
    retries = 3
    for i in range(retries):
        try:
            img.load()
        except OSError as err:
            if i + 1 == retries:
                log(f"WARNING: Loading image failed. [{i + 1}/{retries}]")
                raise err
            else:
                log(f"WARNING: Loading image failed. Trying again... [{i + 1}/{retries}]")
    # unpack data
    e = Image._getencoder(img.mode, 'raw', img.mode)
    e.setimage(img.im)

    # NumPy buffer for the result
    shape, typestr = Image._conv_type_shape(img)
    data = np.empty(shape, dtype=np.dtype(typestr))
    mem = data.data.cast('B', (data.data.nbytes,)) # type: ignore

    bufsize, s, offset = 65536, 0, 0
    while not s:
        l, s, d = e.encode(bufsize)
        mem[offset:offset + len(d)] = d
        offset += len(d)
    if s < 0:
        raise RuntimeError("encoder error %d in tobytes" % s)
    return data


def guessMontageDims(tile_num):
    # Guess based on factors as equal as possible
    i = int(tile_num ** 0.5 + 0.5)
    while tile_num % i != 0:
        i -= 1
    mont_shape = np.array([i, tile_num / i], dtype=int)
    return mont_shape

def serialem_check(func):
    """Decorator to check if SerialEM is running."""

    @wraps(func)
    def inner(*args, **kwargs):
        if "serialem" not in sys.modules:
            raise ModuleNotFoundError("SerialEM is not running!")
        return func(*args, **kwargs)
        
    return inner

def dummy_skip(func):
    """Decorator to skip function in DUMMY mode."""

    @wraps(func)    # wraps is needed to access proper function __name___
    def inner(*args, **kwargs):
        if not config.DUMMY:
            return func(*args, **kwargs)
        else:
            return log(f"WARNING: {func.__module__}.{func.__name__} is skipped in DUMMY mode.")
        
    return inner

def dummy_replace(replacement_func_name):
    """Decorator with argument to take replacement function for function uses in DUMMY mode."""

    def dummy_replace_decorator(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            log(f"DEBUG: {self}, {args}, {kwargs}")

            if not config.DUMMY:
                return func(self, *args, **kwargs)
            else:
                replacement_func = getattr(self, replacement_func_name)
                log(f"WARNING: {func.__module__}.{func.__name__} is replaced with {replacement_func.__module__}.{replacement_func.__name__} in DUMMY mode.")
                return replacement_func(*args, **kwargs)
        return inner
    return dummy_replace_decorator

def saveSettings(file, vars, start="automation_level", end="max_tilt"):
    """Saves SPACEtomo script settings."""

    config = ConfigParser(allow_no_value=True)
    config.optionxform = str
    config["INFO"] = {
        "SPACEtomo version": __version__,
        "Datetime": time.strftime('%d.%m.%Y %H:%M:%S', time.localtime())
    }
    config["SETTINGS"] = {}
    save = False
    for var in vars:                                    # globals() is ordered by creation, start and end points might have to be adjusted if script changes
        if var == start:                                # first var to save
            save = True
        if save:
            config["SETTINGS"][var] = str(vars[var])
        if var == end:                                  # last var to save
            break
    with open(file, "w") as f:
        config.write(f)

def loadSettings(file):
    if file.exists():
        config = ConfigParser()
        config.read(file)
        settings = {}
        for setting, value in config["SETTINGS"].items():
            # Recapitalize name
            if setting.startswith("wg") or setting.startswith("mm") or setting.startswith("im"):
                setting = setting[:2].upper() + setting[2:]
            # Cast to proper type
            settings[setting] = castString(value)
        return settings
    else:
        log("ERROR: Cannot start SPACEtomo because no SPACE settings file was found!")
        sys.exit()

def castString(string):
    """Casts string into int, float, Path, list(recursive), bool or string."""

    if not string:
        return string
    string = string.strip().strip("'\"")
    try:
        var = int(string)
    except ValueError:
        try:
             var = float(string)
        except ValueError:
            if Path(string).exists():
                var = Path(string)
            else:
                if string.startswith("[") and string.endswith("]"):
                    var = [castString(val) for val in string.strip("[]").split(",")]
                    if len(var) == 1 and var[0] == "":
                        var = []
                else:
                    if string.lower() == "true" or string.lower() == "yes" or string.lower() == "y" or string.lower() == "on":
                        var = True
                    elif string.lower() == "false" or string.lower() == "no" or string.lower() == "n" or string.lower() == "off":
                        var = False
                    elif string.lower() == "none":
                        var = None
                    else:
                        var = string
    return var

def waitForFile(file, message, function_call=None, msg_interval=60):
    """Stalls run until file was created."""

    start_time = time.time()
    next_update = start_time + msg_interval

    # Initial function call
    if function_call:
        function_call()

    # Use glob if path contains wildcard *
    if "*" in str(file):
        check = lambda: list(file.parent.glob(file.name))
    else:
        check = lambda: file.exists()

    while not check():
        now = time.time()
        if now > next_update:
            log(message + f" [{round((now - start_time) / 60)} min]")
            if function_call:
                function_call()
            next_update = now + msg_interval
        time.sleep(1)

def monitorExternal(external_dir):
    from SPACEtomo.modules import ext
    unprocessed_mm_list, unprocessed_seg_list, unprocessed_wg_list = ext.monitorFiles(external_dir)
    unprocessed_all = unprocessed_mm_list + unprocessed_seg_list + unprocessed_wg_list
    if unprocessed_all:
        log(f"# Current processing queue:\n" + "\n# ".join(unprocessed_all), color=1)

def writeMrc(file, image, pix_size):
    """Saves image as mrc file."""

    # Mrc cannot handle float64
    if image.dtype == np.float64:
        image = image.astype(np.float32)

    with mrcfile.new(file, overwrite=True) as mrc:
        mrc.set_data(image)
        mrc.voxel_size = (pix_size * 10, pix_size * 10, pix_size * 10)
        mrc.update_header_from_data()
        mrc.update_header_stats()

# Write PACEtomo target file
def writeTargets(targets_file, targets, geo_points=[], saved_run=False, resume={"sec": 0, "pos": 0}, settings=None):
    """Writes tgts file in PACEtomo format."""

    output = ""
    if settings:
        for key, val in settings.items():
            if val != "":
                output += f"_set {key} = {val}\n"
        output += "\n"
    if resume["sec"] > 0 or resume["pos"] > 0:
        output += f"_spos = {resume['sec']},{resume['pos']}\n\n"
    for t, target in enumerate(targets):
        output += f"_tgt = {str(t + 1).zfill(3)}\n"
        for key in target.keys():
            output += f"{key} = {target[key]}\n"
        if saved_run:
            output += "_pbr\n"
            for key in saved_run[t][0].keys():
                output += f"{key} = {saved_run[t][0][key]}\n"
            output += "_nbr\n"
            for key in saved_run[t][1].keys():
                output += f"{key} = {saved_run[t][1][key]}\n"       
        output += "\n"
    for g, geo in enumerate(geo_points):
        output += f"_geo = {g + 1}\n"
        for key in geo.keys():
            output += f"{key} = {geo[key]}\n"
        output += "\n"
    with open(targets_file, "w") as f:
        f.write(output)


def loadDatasetJson(file_path):
    """Loads dataset.json from nnU-Net model."""

    file_path = Path(file_path)
    if file_path.exists():
        with open(file_path, "r") as f:
            dataset_json = json.load(f)

        # Get Categories
        cats = dataset_json["labels"]

        # Get number of images
        img_num = dataset_json["numTraining"]

        # Get pixel size
        if "pixel_size" in dataset_json.keys():
            pix_size = dataset_json["pixel_size"]
        else:
            pix_size = None

        return cats, pix_size, img_num
    else:
        return None, None, None
    

def cyclicRange(start, end, increment):
    """Generates cyclic values."""
    
    while True:
        #for i in range(start, end, increment):
        for i in np.arange(min(start, end), max(start, end), increment):
            yield i

def parseMdoc(mdoc_file):
    """Reads mdoc file into list of item dicts containing all entry values in lists."""
    
    header = ""
    items = []
    
    with open(mdoc_file) as f:
        # Loop over line iterator
        for line in f:
            col = line.strip().split(" ")
            # Skip empty lines
            if not col[0]: 
                continue
            # Instantiate new item on ZValue
            if col[0] == "[ZValue":
                items.append({"ZValue": col[2].strip("]")})

                # Item loop over line iterator
                for entry in f:
                    col = entry.strip().split(" ")
                    # Exit item on empty line
                    if not col[0]:
                        break
                    items[-1][col[0]] = [val for val in col[2:]]
            else:
                header += line

    return header, items
    
def synonymKeys(dictionary: dict, from_list: list, to_list: list):
    """Takes a dictionary and replaces keys in from_list with keys in to_list."""

    for from_key, to_key in zip(from_list, to_list):
        if from_key in dictionary.keys():
            dictionary[to_key] = dictionary[from_key]
            del dictionary[from_key]
    return dictionary

def hashFile(file_path):
    """Gets hash of file contents to monitor changes."""

    if file_path.exists():
        with open(file_path, "rb") as f:
            return hashlib.md5(f.read()).hexdigest()
    else:
        # Empty
        return hashlib.md5(b'\x00').hexdigest()
        
def crossCorrCoeff(img1, img2):
    """Calculates maximum cross correlation coefficient."""

    # Ensure images have same size by center cropping
    if img1.shape != img2.shape:
        img1 = img1[(img1.shape[0] - min(img1.shape[0], img2.shape[0])) // 2: (img1.shape[0] + min(img1.shape[0], img2.shape[0])) // 2, (img1.shape[1] - min(img1.shape[1], img2.shape[1])) // 2, (img1.shape[1] + min(img1.shape[1], img2.shape[1])) // 2]
        img2 = img2[(img2.shape[0] - min(img1.shape[0], img2.shape[0])) // 2: (img2.shape[0] + min(img1.shape[0], img2.shape[0])) // 2, (img2.shape[1] - min(img1.shape[1], img2.shape[1])) // 2, (img2.shape[1] + min(img1.shape[1], img2.shape[1])) // 2]

    image_product = np.fft.fft2(img1) * np.fft.fft2(img2).conj()
    cc = np.fft.fftshift(np.fft.ifft2(image_product))
    return np.max(cc.real)

def getMrcExtendedHeader(file_path):
    """Parses extended header of mrc file and returns data for each section. (At time of implementation the mrcfile 1.5.3 package does not support SERI extended headers.)"""

    with open(file_path, "rb") as mrc_file:
        mrc = mrc_file.read(4096)

    # MRC header format SERI (SERI is not covered by mrcfile)
    format = ""
    for char in struct.iter_unpack("s", mrc[104:108]):
        format += char[0].decode("utf-8")
    log(f"DEBUG: MRC format: {format}")

    if format == "SERI":
        # Get number of sections
        section_number = struct.unpack("i", mrc[8: 12])[0]
        log(f"DEBUG: Number of sections: {section_number}")

        # Get bytes per section
        bytes_per_section = struct.unpack("h", mrc[128: 130])[0]
        log(f"DEBUG: Bytes per section: {bytes_per_section}")

        # Bitflags
        bitflags = struct.unpack("h", mrc[130: 132])[0]
        log(f"DEBUG: Bitflags: {bitflags}")
        # Figure out contained data
        """
        https://bio3d.colorado.edu/imod/doc/mrc_format.txt
        1 = Tilt angle in degrees * 100  (2 bytes)
        2 = X, Y, Z piece coordinates for montage (6 bytes)
        4 = X, Y stage position in microns * 25   (4 bytes)
        8 = Magnification / 100 (2 bytes)
        16 = Intensity * 25000  (2 bytes)
        32 = Exposure dose in e-/A2, a float in 4 bytes
        128, 512: Reserved for 4-byte items
        64, 256, 1024: Reserved for 2-byte items
        If the number of bytes implied by these flags does
        not add up to the value in nint, then nint and nreal
        are interpreted as ints and reals per section
        """
        data_byte_lengths = {"tilt_angle": [2], "piece_coords": [2, 2, 2], "stage": [2, 2], "mag": [2], "intensity": [2], "exposure": [4], "reserved1": [2], "reserved2": [4], "reserved3": [2], "reserved4": [4], "reserved5": [2]}
        data_conversion_factors = {"tilt_angle": 1 / 100, "piece_coords": 1, "stage": 1 / 25, "mag": 100, "intensity": 1 / 25000, "exposure": 1, "reserved1": 1, "reserved2": 1, "reserved3": 1, "reserved4": 1, "reserved5": 1}

        data_bits = reversed([1 if digit == "1" else 0 for digit in bin(bitflags)[2:]])
        data_entries = [entry for entry, bit in zip(data_byte_lengths.keys(), data_bits) if bit]
        log(f"DEBUG: Found data entries: {data_entries}")

        section_data = []
        section_data_raw = []
        # Loop over sections in extended header
        for i in range(1024, 1024 + bytes_per_section * section_number, bytes_per_section): # extended header starts at byte 1024
            section = {}
            # Get bytes for section
            section_raw = mrc[i:i + bytes_per_section]
            # Set pointer to start of section
            start = i
            # Extract data for each entry
            for entry in data_entries:
                # Exposure is formatted as float, rest is short
                byte_type = "h" if entry != "exposure" else "f"
                # Loop over byte lengths and append list if more than one value per entry
                if len(data_byte_lengths[entry]) > 1:
                    section[entry] = []
                    for length in data_byte_lengths[entry]:
                        # Convert bytes to value
                        section[entry].append(struct.unpack(byte_type, mrc[start:start + length])[0] * data_conversion_factors[entry])
                        # Move pointer by read length
                        start += length
                else:
                    # Convert bytes to value
                    section[entry] = struct.unpack(byte_type, mrc[start:start + data_byte_lengths[entry][0]])[0] * data_conversion_factors[entry]
                    # Move pointer by read length
                    start += data_byte_lengths[entry][0]
                    
            section_data.append(section)
            section_data_raw.append(section_raw)

        return section_data, section_data_raw
    
    # TODO: handle other header formats (FEI1 and FEI2 are already implemented in mrcfile)
    return None

def timeoutCall(function, timeout=5, retries=3):
    """Runs function as thread with timeout and retry limitations."""

    for attempt in range(retries):
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(function)
            try:
                # Wait up to timeout for the function to complete
                result = future.result(timeout=timeout)
                log(f"DEBUG: Attempt {attempt + 1} was successful!")
                return result
            except concurrent.futures.TimeoutError:
                log(f"WARNING: Call to {function.__name__} timed out! Retrying...")
                # Cancel the future
                future.cancel()
                time.sleep(1)
    raise concurrent.futures.TimeoutError(f"Call to {function.__name__} timed out {retries} times!")

def guiProcess(gui_type, file_path="", auto_close=False):
    """Starts new process to call GUI."""

    file_path = str(file_path)

    args = [gui_type, file_path]
    if auto_close:
        args.append("--auto_close")

    log(f"DEBUG: GUI args: {args}")

    DETACHED_PROCESS = 0x00000008 # From here: https://stackoverflow.com/questions/89228/calling-an-external-command-in-python#2251026
    try:
        subprocess.Popen([sys.executable, Path(__file__).parent.parent / "GUI.py", *args], creationflags=DETACHED_PROCESS)
    except ValueError:      # Creationflags only supported on Windows
        subprocess.Popen([sys.executable, Path(__file__).parent.parent / "GUI.py", *args])


#!Python
# ===================================================================
#ScriptName SPACEtomo
# Purpose:      Finds lamellae in grid map, collects lamella montages and finds targets using deep learning models.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/05/31
# Revision:     v1.2
# Last Change:  2024/11/20: switched to nav bases item shift, added wait for threaded saving (temporary)
#               2024/10/21: Added retention of high confidence lamellae
#               2024/09/04: converted settings to configparser
#               2024/08/19: finished refactor (except ext.py)
#               2024/08/16: completed level 1 refactor, completed level 2 refactor, started refactor of target setup
#               2024/08/13: started switching to Microscope and Navigator classes, also started switch to Pathlib
#               2024/04/12: fixed grid_list conversion
#               2024/04/09: fixed GUI opening multiple times, rename preexisting run file, moved import functions past SetDirectory
#               2024/04/02: moved Rec param saving before MM montages to allow earlier target picking, added setting scripts for Acquire at Items, added status line for multi grid
#               2024/03/26: added check for GPU
#               2024/03/25: fixes after Krios 3 test, added opening of SPACEtomo_tgt, added manual_selection by creating empty segmentation, added more text outputs
#               2024/03/18: added multi-grid support, added waiting for manual inspection (x_inspected.txt file)
#               2024/03/08: outsourced target selection to updateQueue, introduced TgtParams
#               2024/03/07: adjusted for external target selection
#               2024/02/14: adjustments to run lamella detection externally
#               2024/02/13: split collectWGMap from lamella detection (1.1)
#               2023/12/12: added version check
#               2023/11/06: used IM lamella detection as confirmation
#               2023/10/31: fixes after Krios test
#               2023/10/27: adjusted plotting option
#               2023/10/25: added open nav file
#               2023/10/23: added extra tracking
#               2023/10/16: fixes after Krios test
#               2023/10/12: fixes after Krios test
#               2023/10/05: added option for inference on external machine
#               2023/10/04: outsourced functions
#               2023/10/03: added auto identification of geo points, added geo points to tgts fie
#               2023/10/02: fixes after Krios test, added intermediate mag lamella detection for offset refinement
#               2023/09/27: clean up, added manual intervention, montage fixes
#               2023/09/26: fixes after Krios test, added queue for SPACEruns
#               2023/09/22: switched to nnUNet model and adjusted script accordingly including target selection from segmentation
#               2023/07/13: added writing of SPACE_run file to be read by post script
#               2023/07/10: binned heat maps, fixed target specimen shifts, added score of targets based on camera dimensions, cleanup
#               2023/07/07: added target setup, tested and fixed on Krios
#               2023/07/05: fixes and cleanup, added per lamella montage size
#               2023/07/04: first Krios test
#               2023/06/19: added montage loop
#               2023/05/31: first test in SerialEM
# ===================================================================

import SPACEtomo.config as config

# Put SerialEM python modules on PATH
import sys
if not config.DUMMY:
    sys.path.insert(len(sys.path), config.SERIALEM_PYTHON_PATH)
    try:
        import serialem as sem
        SERIALEM = True

        # Connect to SEM
        from SPACEtomo.modules.utils_sem import connectSerialEM
        connectSerialEM(config.SERIALEM_IP, config.SERIALEM_PORT)

    except ModuleNotFoundError:
        SERIALEM = False
else:
    SERIALEM = False
    
import time
from copy import deepcopy
from pathlib import Path
import SPACEtomo.modules.ext as space_ext
from SPACEtomo.modules.utils import log
from SPACEtomo.modules.scope import Microscope, ImagingParams
from SPACEtomo.modules.nav import Navigator
from SPACEtomo.modules.buf import Buffer
from SPACEtomo.modules.mod_wg import WGModel, Boxes
from SPACEtomo.modules.tgt import PACEArea
import SPACEtomo.modules.utils as utils
import SPACEtomo.modules.utils_sem as usem

if config.DUMMY:    # Shadow classes with dummy classes
    log(f"WARNING: Running SPACEtomo in DUMMY mode! Run SPACEconfig dummy command to toggle it off!")
    from SPACEtomo.modules.dummy.scope import MicroscopeDummy as Microscope
    from SPACEtomo.modules.dummy.buf import BufferDummy as Buffer
    import SPACEtomo.modules.dummy.utils_sem as usem

### Run on import: Connect to SerialEM, get session directory

# Connect SerialEM
usem.connectSerialEM(config.SERIALEM_IP, config.SERIALEM_PORT)

# Check SerialEM version
usem.checkVersion()

# Check if session dir has been chosen previously or ask for it
SES_DIR = usem.getSessionDir()

###

def main():
    # Read settings file
    settings = utils.loadSettings(SES_DIR / "SPACEtomo_settings.ini")
    if settings["manual_selection"] and not settings["MM_wait_for_inspection"]: # Force inspection in case of manual selection
        settings["MM_wait_for_inspection"] = True

    # Check model
    if config.NO_MM_MODEL:
        log(f"ERROR: Lamella segmentation model could not be found! Please import it using the SPACEmodel command!")
        return
    
    # Check multi-grid status
    remaining_grid_list = usem.getGridList(settings["grid_list"], settings["automation_level"])

    # Show warning if grid will be changed after target setup
    if settings["automation_level"] == 4 and len(remaining_grid_list) > 1:
        log(f"WARNING: You selected multiple grids in combination with target selection. Reloading grids might cause problems with realignment to target areas.")
        usem.confirmationBox(f"WARNING: You selected {len(remaining_grid_list)} grids in combination with target selection. Reloading grids might cause problems with realignment to target areas!")

    # Initialize microscope control
    imaging_params = ImagingParams(settings["WG_image_state"], settings["IM_mag_index"], settings["MM_image_state"], file_dir=SES_DIR)
    microscope = Microscope(imaging_params)

    # Check if imaging states are valid
    log(f"Checking imaging states...")
    if not microscope.changeImagingState(imaging_params.MM_image_state, low_dose_expected=True):
        log(f"ERROR: The provided medium mag imaging state [MM_image_state] is not in Low Dose mode!")
        return

    # Run on every grid
    for grid_slot in remaining_grid_list:

        # Get grid name
        grid_name = microscope.autoloader[grid_slot]

        # Setup dir
        CUR_DIR, MAP_DIR, EXTERNAL_RUN = usem.prepareEnvironment(SES_DIR, grid_name, settings["external_map_dir"])

        # Save settings for SPACEtomo_postAction.py
        utils.saveSettings(MAP_DIR / "SPACEtomo_settings.ini", settings)
        utils.saveSettings(CUR_DIR / "SPACEtomo_settings.ini", settings)

    ########################################
    ############ STEP 1: WG map ############
    ########################################

        if settings["automation_level"] >= 1:

            # Load model
            WG_model = WGModel(EXTERNAL_RUN)  # load a custom model

            # Load grid
            nav = Navigator(CUR_DIR / (grid_name + ".nav"))
            Buffer.nav = nav            # Provide reference for all buffers

            if config.DUMMY:            # Provide additional references to dummy classes
                microscope.cur_dir = CUR_DIR
                microscope.map_dir = MAP_DIR
                microscope.nav = nav
                Buffer.imaging_params = imaging_params
                if imaging_params.rec_ta_rotation is None:
                    log(f"ERROR: Cannot run in DUMMY mode without mic_params.json!")
                    return

            microscope.loadGrid(grid_slot)
            if config.aperture_control:
                microscope.changeObjAperture(0)
            microscope.openValves()

            # Check for WG montage
            wg_nav_id = nav.getIDfromNote(grid_name + ".mrc")
            if wg_nav_id is None:
                # Make WG montage
                wg_nav_id = microscope.collectFullMontage(WG_model, overlap=config.WG_montage_overlap)
                nav.pull()

            # Save WG map
            wg_map = Buffer(nav_id=wg_nav_id)
            save_future = wg_map.saveImg(MAP_DIR / (grid_name + "_wg.png"), WG_model.pix_size)

            # Find lamellae on grid
            box_file = MAP_DIR / (grid_name + "_wg_boxes.json")
            if not box_file.exists():
                if not EXTERNAL_RUN:
                    if save_future is not None:
                        save_future.result()
                        save_future = None
                    WG_model.findLamellae(MAP_DIR, grid_name, threshold=config.WG_detection_threshold, device=utils.DEVICE)
                else:
                    # Wait for boxes file to be written
                    log("Waiting for external lamella detection...")
                    utils.waitForFile(box_file, "WARNING: Still waiting for lamella detection. Check if SPACEtomo_monitor.py is running!")
            else:
                log(f"WARNING: Previously detected lamellae were found. Skipping lamella detection!")

            # Open lamella selection GUI if wait_for_inspection is selected and maps are not external and not already inspected
            if settings["WG_wait_for_inspection"] and not EXTERNAL_RUN and not (MAP_DIR / (grid_name + "_wg_inspected.txt")).exists():
                utils.guiProcess("lamella", MAP_DIR / (grid_name + "_wg.png"), auto_close=True)

            # Wait for lamella boxes to be inspected
            if settings["WG_wait_for_inspection"]:
                log("Waiting for lamella detection to be inspected by operator...")
                inspect_file = MAP_DIR / (grid_name + "_wg_inspected.txt")
                utils.waitForFile(inspect_file, "Still waiting for inspection by operator...", msg_interval=180)

            # Load boxes from file and remove undesired classes
            lamella_boxes = Boxes(box_file, label_prefix="PL", exclude_cats=settings["exclude_lamella_classes"])

            # Check for previously determined lamella positions
            lamella_PL_ids = nav.searchByEntry("label", "PL", partial=True)     # Preliminary lamellae (PL)
            lamella_FL_ids = nav.searchByEntry("label", "FL", partial=True)     # Final lamellae (FL)
            log(f"DEBUG: PL entries: {len(lamella_PL_ids)}")
            log(f"DEBUG: FL entries: {len(lamella_FL_ids)}")

            # Add boxes to nav via map buffer (if no lamella in nav yet)
            if len(lamella_PL_ids) == 0:
                wg_map.addNavBoxes(lamella_boxes, padding_factor=config.MM_padding_factor)
                nav.pull()

            # Check if intermediate mag step was done (if there are as many images as preliminary lamella points and final lamella points)
            log(f"DEBUG: Test if IM has to be done:\nIM pngs: {len(list(MAP_DIR.glob(grid_name + '_IM*_wg.png')))}\nPL entries: {len(lamella_PL_ids)}\nFL entries: {len(lamella_FL_ids)}\n\n")
            if not len(list(MAP_DIR.glob(grid_name + "_IM*_wg.png"))) == len(lamella_PL_ids) or not lamella_FL_ids:
                log("DEBUG: Collecting IM images...\n")
                microscope.changeImagingState(settings["WG_image_state"], low_dose_expected=False)
                microscope.setMag(settings["IM_mag_index"])
                lamella_nav_ids = nav.searchByEntry("label", "PL", partial=True)     # Preliminary lamellae (PL)
                shifted = False
                eucentricity = False
                for n, nav_id in enumerate(lamella_nav_ids):

                    # If lamella nav item was already converted from polygon to map, skip it
                    if nav.items[nav_id].entries["Type"] == ["2"]:
                        log(f"WARNING: Lamella {n + 1} was already imaged at intermediate mag and will be skipped.")
                        continue

                    log(f"Moving to lamella {n + 1}...")
                    microscope.moveStage(nav.items[nav_id].stage)

                    # Go to eucentricity roughly at IM mag
                    if not eucentricity:
                        microscope.eucentricity()
                        eucentricity = True
                        # Update z for all lamella items
                        for nid in lamella_nav_ids:
                            nav.items[nid].changeZ(microscope.stage[2])
                        nav.push()

                    # Collect IM image and save as map
                    log(f"Collecting intermediate mag image...")
                    im_file = CUR_DIR / (grid_name + f"_IM{n + 1}.mrc")
                    microscope.record(save=im_file)
                    im_img = Buffer(buffer="A")
                    im_nav_id = nav.newMap(buffer=im_img, img_file=im_file, label=f"PL{n + 1}", note=nav.items[nav_id].note, coords=nav.items[nav_id].stage) # Coords are only needed for DUMMY and ignored for normal map

                    # Update IM pix_size
                    if not imaging_params.IM_pix_size:
                        imaging_params.IM_pix_size = im_img.pix_size

                    # Save IM image
                    if save_future is not None:
                        save_future.result()
                        save_future = None
                    save_future = im_img.saveImg(MAP_DIR / (grid_name + f"_IM{n + 1}_wg.png"), WG_model.pix_size)

                    # Retain preliminary polygon nav item in case no lamella is detected at IM
                    pre_item = deepcopy(nav.items[nav_id])

                    # Overwrite polygon with map
                    nav.replaceItem(nav_id, im_nav_id)
                    im_img.nav_id = nav_id # Update nav id of buffer

                    # Find lamellae on image
                    log(f"Finding lamella...")
                    box_file = MAP_DIR / (grid_name + f"_IM{n + 1}_wg_boxes.json")
                    if not EXTERNAL_RUN:
                        if save_future is not None:
                            save_future.result()
                            save_future = None
                        WG_model.findLamellae(MAP_DIR, grid_name + f"_IM{n + 1}", threshold=config.WG_detection_threshold, device=utils.DEVICE)
                    else:
                        # Wait for boxes file to be written
                        log("Waiting for external lamella detection...")
                        utils.waitForFile(box_file, "WARNING: Still waiting for lamella detection. Check if SPACEmonitor is running!")

                    im_boxes = Boxes(box_file)

                    if im_boxes:
                        if len(im_boxes) > 1:
                            log(f"WARNING: Found more than one lamella. Using lamella closest to image center.")

                        # Only keep box closest to center
                        im_boxes.sortBy("center_dist")
                        im_boxes.boxes = im_boxes.boxes[:1]

                        log(f"NOTE: Detected lamella of class: {config.WG_model_categories[im_boxes.boxes[0].cat]} [{round(im_boxes.boxes[0].prob * 100)} %]")

                        # Check if too close to previously detected lamella
                        lamella_FL_ids = nav.searchByEntry("label", "FL", partial=True)     # Final lamellae (FL)
                        log(f"DEBUG: FL ids: {lamella_FL_ids}")
                        if nav.searchByCoords(im_img.px2stage(im_boxes.boxes[0].center * (im_boxes.pix_size / im_img.pix_size)), margin=settings["WG_distance_threshold"], subset=lamella_FL_ids):
                            log(f"DEBUG: Distance: {im_img.px2stage(im_boxes.boxes[0].center * (im_boxes.pix_size / im_img.pix_size))}")
                            log(f"DEBUG: Search result: {nav.searchByCoords(im_img.px2stage(im_boxes.boxes[0].center * (im_boxes.pix_size / im_img.pix_size)), margin=settings['WG_distance_threshold'], subset=lamella_FL_ids)}")
                            log(f"WARNING: Lamella seems to be among already detected lamellae and will be skipped.")
                            continue

                        # Add box to nav via map buffer
                        im_img.addNavBoxes(im_boxes, labels=[f"FL{n + 1}"], padding_factor=config.MM_padding_factor)                 # Add final lamella
                        nav.pull()

                        # Shift items by vector between new box center stage coords and nav item stage coords
                        if not shifted:
                            #nav.shiftItems(nav.items[nav_id].getVector(im_img.px2stage(im_boxes.boxes[0].center * (im_boxes.pix_size / im_img.pix_size))))
                            shift = nav.items[nav_id].getVector(im_img.px2stage(im_boxes.boxes[0].center * (im_boxes.pix_size / im_img.pix_size)))

                            for remaining_nav_id in lamella_nav_ids[n + 1:]:
                                log(f"DEBUG: Shifted item {remaining_nav_id} by {shift}")
                                nav.items[remaining_nav_id].changeStage(shift, relative=True)

                            shifted = True
                            nav.push()

                    else:
                        # Retain preliminary lamella if initial confidence was high (includes manually selected lamella)
                        if float(pre_item.entries["UserValue1"][0]) >= 0.9:
                            nav.items.append(pre_item) # Add initial polygon back to nav
                            pre_item.nav_index = len(nav) # Adjust nav_index
                            nav.push()
                            # Only rename label and note after nav was pushed as it will update SerialEM directly
                            pre_item.changeLabel(f"FL{n + 1}")
                            pre_item.changeNote(f"FL{n + 1}: " + pre_item.note.split(":")[1])
                            log(f"WARNING: No lamella detected! Initially detected lamella was retained due to high confidence [{round(float(pre_item.entries['UserValue1'][0]) * 100)} %].")
                        else:
                            log(f"WARNING: No lamella detected. If you want to add lamellae manually, please select wait_for_inspection in the settings.")

            lamella_nav_ids = nav.searchByEntry("label", "FL", partial=True)     # Final lamellae (FL)

            if not lamella_nav_ids:
                log(f"WARNING: No lamellae found on grid [{grid_name}]. If lamellae are visually identifiable, please check your settings or continue manually.")
                continue    # to next grid

            log("##### Completed lamella detection step! [Level 1] #####")

    #########################################
    ############ STEP 2: MM maps ############
    #########################################

        if settings["automation_level"] >= 2:

            # Load model
            MM_model = space_ext.MMModel()

            # Insert objective aperture
            if config.aperture_control:
                microscope.changeObjAperture(config.objective_aperture)

            # Enter Low Dose Mode
            microscope.changeImagingState(settings["MM_image_state"], low_dose_expected=True)
            if not microscope.low_dose:
                log("WARNING: Lamella montage imaging state is NOT in Low Dose! This will cause issues with target selection.")

            # Get imaging parameters and save
            microscope.changeLowDoseArea("V")
            imaging_params.getViewParams()
            microscope.changeLowDoseArea("R")
            imaging_params.getRecParams()

            MM_model.setDimensions(imaging_params)

            # Set up common parameters
            tgt_params = space_ext.TgtParams(settings["target_list"], settings["avoid_list"], MM_model, settings["sparse_targets"], settings["target_edge"], settings["penalty_weight"], settings["target_score_threshold"], settings["max_tilt"], imaging_params, settings["extra_tracking"], config.max_iterations)

            # Export params for postAction script
            imaging_params.export(CUR_DIR)
            tgt_params.export(CUR_DIR)
            imaging_params.export(MAP_DIR)
            tgt_params.export(MAP_DIR)

            # Remove preexisting SPACE_runs
            if not EXTERNAL_RUN and (MAP_DIR / "SPACE_runs.json").exists():
                (MAP_DIR / "SPACE_runs.json").replace(MAP_DIR / "SPACE_runs.json~")

            # Collect MM montages at all lamella positions
            MM_map_ids = []
            for i, nav_id in enumerate(lamella_nav_ids):
                # Make sure no MM map is waiting in queue before starting next montage. TODO: better queue management
                if settings["automation_level"] >= 3 and not EXTERNAL_RUN and i > 0:
                    space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params)

                # Get lamella map name
                map_name = f"{grid_name}_L{str(i + 1).zfill(2)}"
                map_file = CUR_DIR / (map_name + ".mrc")

                # Collect montage
                log(f"Collecting map for lamella {map_name}...")

                # Check if montage already exists in nav
                map_id = nav.getIDfromNote(map_file.name, warn=False)

                # Check if file also exists
                if not map_file.exists():
                    # Check if it was marked for reacquisition by renaming and update nav note accordingly
                    if old_map_files := list(map_file.parent.glob(map_file.stem + "_old*.mrc")):
                        nav.items[map_id].changeNote(old_map_files[-1].name)
                        map_id = None
                        # Reset inspection status
                        if (MAP_DIR / (map_file.stem + "_inspected.txt")).exists():
                            (MAP_DIR / (map_file.stem + "_inspected.txt")).unlink()

                # Collect map if no nav item found
                if map_id is None:
                    # Delete potentially aborted montage file
                    if map_file.exists():
                        map_file.unlink()
                    map_id = microscope.collectPolygonMontage(nav_id, map_file, config.MM_montage_overlap)
                    nav.pull()
                    nav.items[map_id].changeLabel(f"L{str(i + 1).zfill(2)}")
                MM_map_ids.append(map_id)

                # Save montage as rescaled input image
                log(f"Saving map for lamella {map_name}...")
                map_img = Buffer(nav_id=map_id)
                save_future = map_img.saveImg(MAP_DIR / (map_file.stem + ".png"), MM_model.pix_size)
                if save_future is not None:
                    save_future.result() # Temporary, because montage collection is not threaded, TODO: multiprocessing instead
                    save_future = None

                if settings["automation_level"] >= 3:
                    if settings["manual_selection"]:
                        MM_model.saveEmptySeg(MAP_DIR, map_name)
                    else:
                        if not EXTERNAL_RUN:
                            # Queue SPACEtomo run for new montage
                            space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params)

                # Open tgt selection GUI if manual_selection is selected and maps are not external (and only after first map finished collection)
                if settings["automation_level"] >= 4 and i == 0 and settings["manual_selection"] and not EXTERNAL_RUN:
                    utils.guiProcess("targets", MAP_DIR, auto_close=True)
            log("##### Completed collection of lamella maps! [Level 2] #####")

    ##############################################
    ############ STEP 3: Target setup ############
    ##############################################

        total_targets = 0
        if settings["automation_level"] >= 4:
            # Make sure at least the first target selection is ready for setup
            nav.pull()
            map_file = Path(nav.items[MM_map_ids[0]].note)
            utils.waitForFile(MAP_DIR / (map_file.stem + "_points*.json"), 
                            "Waiting for first lamella to be processed before setting up targets...", 
                            function_call=lambda: space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params) if not EXTERNAL_RUN else lambda: utils.monitorExternal(MAP_DIR)
                            )
            # Make sure targets were inspected
            if settings["MM_wait_for_inspection"]:
                # Open GUI if not already opened for manual selection
                if not settings["manual_selection"] and not EXTERNAL_RUN:
                    utils.guiProcess("targets", MAP_DIR, auto_close=True)
                # Wait for inspection
                utils.waitForFile(MAP_DIR / (map_file.stem + "_inspected.txt"), 
                                "Waiting for targets to be inspected by operator...", 
                                function_call=lambda: space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params) if not EXTERNAL_RUN else lambda: utils.monitorExternal(MAP_DIR)
                                )

            # Loop over all MM maps
            for m, map_id in enumerate(MM_map_ids):
                map_name = Path(nav.items[map_id].note).stem
                map_file = MAP_DIR / (map_name + ".png")

                # Check if file exists
                if not (CUR_DIR / (map_name + ".mrc")).exists():
                    # Check if it was marked for reacquisition by renaming and update nav note accordingly
                    if old_map_files := list(CUR_DIR.glob(map_name + "_old*.mrc")):
                        nav.items[map_id].changeNote(old_map_files[-1].name)
                        log(f"WARNING: {map_name} was marked for reacquisition and target setup will be skipped.")
                        continue

                map_img = Buffer(nav_id=map_id)

                if not EXTERNAL_RUN:
                    space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params)

                # Check if point file exists
                if not list(MAP_DIR.glob(map_name + "_points*.json")) or (settings["MM_wait_for_inspection"] and not (MAP_DIR / (map_name + "_inspected.txt")).exists()):

                    # Check if there are already items to be acquired
                    nav.pull()
                    num_acq = len(nav.searchByEntry("Acquire", ["1"]))

                    # If acquisition cannot start yet or automation is 4, wait for point files
                    if settings["automation_level"] < 5 or (settings["automation_level"] >= 5 and num_acq == 0):
                        # Wait for points files
                        utils.waitForFile(MAP_DIR / (map_file.stem + "_points*.json"), 
                                        "Waiting for first lamella to be processed before setting up targets...", 
                                        function_call=lambda: space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params) if not EXTERNAL_RUN else None
                                        )
                        # Make sure targets were inspected
                        if settings["MM_wait_for_inspection"]:
                            utils.waitForFile(MAP_DIR / (map_file.stem + "_inspected.txt"), 
                                            "Waiting for targets to be inspected by operator...", 
                                            function_call=lambda: space_ext.updateQueue(MAP_DIR, WG_model, MM_model, imaging_params, tgt_params) if not EXTERNAL_RUN else None
                                            )
                    else:
                        log("Target setup for remaining lamellae postponed until predictions are available.")
                        break           # rely on SPACEtomo_postAction script to setup more targets between PACEtomo runs


                log(f"\nSetting up targets for {map_name}...")

                # Save targets for PACEtomo
                if list(CUR_DIR.glob(map_name + "*_tgts.txt")):
                    log(f"WARNING: Targets file for {map_name} already exists! Skipping target setup!")
                else:
                    target_num = 0

                    # Go over all target areas on lamella
                    point_files = sorted(MAP_DIR.glob(map_name + "_points*.json"))
                    for ta, point_file in enumerate(point_files):
                        if len(point_files) > 1:
                            area_name = map_name + f"_A{ta + 1}"
                        else:
                            area_name = map_name

                        # Instantiate target area
                        target_area = PACEArea(point_file, MM_model, imaging_params, map_img)
                        if len(target_area) == 0: 
                            log(f"WARNING: No targets selected on {area_name}. Skipping...")
                            continue
                        
                        target_area.checkISLimit()

                        # Set up navigator and create virtual maps
                        target_area.prepareForAcquisition(CUR_DIR / (area_name + ".mrc"), nav)

                        # Save target file
                        targets, geo_points, tgt_settings = target_area.getTargetsInfo(area_name)
                        utils.writeTargets(CUR_DIR / (area_name + "_tgts.txt"), targets, geo_points, settings=tgt_settings)
                        target_num += len(targets)

                    log(f"NOTE: Targets selected on {map_name}: {target_num}")

            log("##### Completed target selection step! [Level 4] #####")

    #############################################
    ############ STEP 4: Acquisition ############
    #############################################

    # Start acquire at items
    if settings["automation_level"] >= 5:
        log("The SPACEtomo setup script completed!")
        log(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))

        usem.setupAcquisition(*settings["script_numbers"])
    else:
        usem.exitSPACEtomo()

if __name__ == "__main__":
    main()


# LisC filter with lamella alignment
# without plotting

import os
import time
import numpy as np
from PIL import Image
Image.MAX_IMAGE_PIXELS = None           # removes limit for large images
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from skimage import transform, filters, registration, exposure
from scipy.ndimage import gaussian_filter

PLOT = True

file_name = "/Users/eifabian/Documents/temp/Yeast8_L01_bin4.png"
pix_size = 18.9 * 4 #[nm/px]
tilt_axis = 0 #deg

highpass_threshold = 10000 #[nm]
exposure_crop = (0.7, 1.3)

# Import png image
image = np.array(Image.open(file_name))

# Only consider single channel
if len(image.shape) > 2:
    image = image[:, :, 0]

start = time.time()

print("Image")
plt.figure(figsize=(10,10))
plt.imshow(image, cmap="grey")
plt.axis("off")
plt.show()

ori_dims = image.shape


# FIND ROTATION

# Generate reference
ref_size = (400, 400)
ref = np.zeros(ref_size)
ref[int(ref_size[0] * 0.2): int(ref_size[0] * 0.8), :] = 1

if PLOT:
    print("Reference")
    plt.figure(figsize=(10,10))
    plt.imshow(ref)
    plt.axis("off")
    plt.show()

# Image resize to fit ref
image_sq = image[(image.shape[0] - min(image.shape)) // 2: (image.shape[0] + min(image.shape)) // 2, (image.shape[1] - min(image.shape)) // 2: (image.shape[1] + min(image.shape)) // 2]
image_sq = transform.resize(image_sq, ref_size, anti_aliasing=True)

# Apply soft edge mask
image_filt = image_sq * filters.window("hann", image_sq.shape)
ref_filt = ref * filters.window("hann", ref.shape)

if PLOT:
    print("Image resized to reference with soft edge")
    fig, ax = plt.subplots(1, 2, figsize=(10,10))
    ax[0].imshow(image_filt)
    ax[0].axis("off")
    ax[1].imshow(ref_filt)
    plt.show()

# Fourier transform
fft_img = abs(np.fft.fftshift(np.fft.fft2(image_filt)))
fft_ref = abs(np.fft.fftshift(np.fft.fft2(ref_filt)))

if PLOT:
    print("Fourier transform")
    fig, ax = plt.subplots(1, 2, figsize=(10,10))
    ax[0].imshow(np.log(abs(fft_img)))
    ax[0].axis("off")
    ax[1].imshow(np.log(abs(fft_ref)))
    ax[1].axis("off")
    plt.show()

# High-pass filter
yy = np.linspace(-np.pi / 2, np.pi / 2, fft_img.shape[0])[:, None]
xx = np.linspace(-np.pi / 2, np.pi / 2, fft_img.shape[1])[None, :]
rads = np.sqrt(xx ** 2 + yy ** 2)
filt = 1 - np.cos(rads) ** 2
filt[np.abs(rads) > np.pi / 2] = 1

fft_img = fft_img * filt
fft_ref = fft_ref * filt

# Log-polar conversion
lp_img = transform.warp_polar(fft_img, scaling="log", output_shape=(360,500))
lp_ref = transform.warp_polar(fft_ref, scaling="log", output_shape=(360,500))

if PLOT:
    print("Log-polar converted FFT")
    fig, ax = plt.subplots(1, 2, figsize=(10,10))
    ax[0].imshow(np.log(abs(lp_img)))
    ax[0].axis("off")
    ax[1].imshow(np.log(abs(lp_ref)))
    ax[1].axis("off")
    plt.show()

# Phase correlation (skimage implementation)
shifts, *_ = registration.phase_cross_correlation(lp_img, lp_ref)
#print(shifts)

# Calculate scale
base = np.exp(np.log(lp_img.shape[1] / 2) / lp_img.shape[1])
scale = base ** shifts[1]
print("Scale:", scale)

# Calc rotation
rotation = shifts[0]# - 180
print("Rotation:", rotation)

# Apply rotation
image_rot = transform.rotate(image, rotation, resize=True)

if PLOT:
    plt.figure(figsize=(10,10))
    plt.imshow(image_rot)
    plt.axis("off")
    plt.show()

print("Time to find rotation: ", time.time() - start, " s")

### START FILTERING ###

filter_threshold = highpass_threshold / pix_size
sigma_masks = 6.5 / pix_size

(w, h) = image_rot.shape
half_w, half_h = int(w/2), int(h/2)

# Fourier transform
fft = np.fft.fftshift(np.fft.fft2(image_rot))

if PLOT:
    plt.figure(figsize=(10,10))
    plt.imshow(np.log(abs(fft)))
    plt.axis("off")
    plt.show()

# high pass filter
filter_radius = pix_size * w / highpass_threshold # fourier pixels
print("Filter radius (fourier pixels):", filter_radius, filter_threshold)

mask = np.ones(fft.shape)
mask[:, range(half_h - int(filter_radius), half_h + int(filter_radius))] = 0
mask = gaussian_filter(mask, filter_radius / 4)

fft *= mask

if PLOT:
    plt.figure(figsize=(10,10))
    plt.imshow(np.log(abs(fft)))
    plt.axis("off")
    plt.show()

#highpassed = fftpack.ifft2(fftpack.ifftshift(fft)).real
highpassed = np.fft.ifft2(np.fft.ifftshift(fft)).real

if PLOT:
    plt.figure(figsize=(10,10))
    plt.imshow(highpassed, cmap="grey")
    plt.axis("off")
    plt.show()

highpassed_crop = highpassed[int(exposure_crop[0] * half_w):int(exposure_crop[1] * half_w), int(exposure_crop[0] * half_h):int(exposure_crop[1] * half_h)]
#in_range = (np.mean(highpassed_crop) - np.std(highpassed_crop), np.mean(highpassed_crop) + np.std(highpassed_crop))
#print(in_range)
in_range = (np.min(highpassed_crop), np.max(highpassed_crop))
print("Highpass exposure in-range:", in_range)
highpassed = exposure.rescale_intensity(highpassed, in_range=in_range, out_range=(0, 1))

#plt.figure(figsize=(10,10))
#plt.hist(highpassed)
#plt.show()


# Vacuum mask

lowpassed = gaussian_filter(image_rot, sigma_masks)

if PLOT:
    fig, ax = plt.subplots(figsize=(10,10))
    plt.imshow(lowpassed, cmap="grey")
    rect = Rectangle((exposure_crop[0] * half_h, exposure_crop[0] * half_w), (exposure_crop[1] - exposure_crop[0]) * half_h, (exposure_crop[1] - exposure_crop[0]) * half_w, fill=False, linewidth=1, edgecolor="r")
    ax.add_patch(rect)
    plt.axis('off')
    plt.show()


# Crop to get histogram of lamella area
lowpassed_crop = lowpassed[int(exposure_crop[0] * half_w):int(exposure_crop[1] * half_w), int(exposure_crop[0] * half_h):int(exposure_crop[1] * half_h)]

# Create masks for vacuum and contamination
threshold = np.mean(lowpassed) + 1.5 * np.std(lowpassed)
print("Vac threshold:", threshold)
mask_vacuum = np.ones(image_rot.shape)
mask_vacuum[lowpassed > threshold] = 0


threshold = np.mean(lowpassed_crop) - 1.5 * np.std(lowpassed_crop)
print("Contamination threshold:", threshold)
mask_dark = np.ones(image_rot.shape)
mask_dark[lowpassed < threshold] = 0

# Combine masks
mask_both = (mask_vacuum * mask_dark).astype(bool)

if PLOT:
    plt.figure(figsize=(10,10))
    plt.imshow(mask_both, cmap="grey")
    plt.axis('off')
    plt.show()

# Create final image
final_rot = mask_both * highpassed + np.invert(mask_both) * lowpassed

# Rotate back to original
final = transform.rotate(final_rot, -rotation, resize=True)
final = final[(final.shape[0] - ori_dims[0]) // 2: (final.shape[0] + ori_dims[0]) // 2, (final.shape[1] - ori_dims[1]) // 2: (final.shape[1] + ori_dims[1]) // 2]

plt.figure(figsize=(10,10))
plt.imshow(final, cmap="grey")
plt.axis('off')
plt.show()

print("Total time elapsed:", (time.time() - start), "seconds")

image_save = Image.fromarray(np.uint8(final * 255))
image_save.save(os.path.splitext(file_name)[0] + "_lisc.png")